import asyncio
import websockets
import time
import functools
import json
import uuid
import traceback
import colorama

from typing import *
from dataclasses import asdict

from maica import mtools
from maica.maica_nows import NoWsCoroutine
from maica.maica_utils import *

_CONNS_LIST = ['auth_pool', 'maica_pool', 'mcore_conn', 'mfocus_conn', 'mvista_conn', 'mnerve_conn']

async def drain_buffer(websocket, id: int):
    """Drains a buffer and send through ws."""
    buffer: mtools.StreamBuffer = mtools.buffer_dict.get(id)
    if not buffer:
        await messenger(websocket, 'maica_reconn_buffer_empty', f"Reconnection buffer not present for user id {id}.", '204')
        return

    sent = 0
    async for ws_tuple in buffer:
        sent += 1
        await websocket.send(wrap_ws_formatter(*ws_tuple))

    await messenger(websocket, 'maica_reconn_buffer_drained', f"Reconnection buffer drained for user id {id}, {sent} packets sent.", '200', type=MsgType.INFO)

class BufferedMessenger():
    """As name."""
    @staticmethod
    def _dummy_messenger(websocket=None, *args, **kwargs):
        return sync_messenger(*args, **kwargs)
    
    def __init__(self, id: int):
        """Unlinks previous buffer on initialization."""
        self.id = id
        self.wsint = False
        self.we = None

        mtools.buffer_dict.del_id(id)

    async def __call__(self, *args, **kwargs):
        if not self.wsint:
            try:
                return await messenger(*args, **kwargs)
            except websockets.WebSocketException as we:
                self.wsint = True
                self.we = we
                sync_messenger(info="<WSINT, storing remaining to buffer>", type=MsgType.PLAIN, color=colorama.Fore.LIGHTYELLOW_EX)

                mtools.buffer_dict.add_id(self.id)
                self._buffer: mtools.StreamBuffer = mtools.buffer_dict[self.id]
                self._timestamp = self._buffer.timestamp

        if self.wsint:
            ws_tuple = self._dummy_messenger(*args, **kwargs)
            await self._buffer.aappend(ws_tuple)

    async def seal(self):
        """
        Announce corresponding buffer exhausted.
        Also raises ws exception if captured, remains the destroying procedure intact.
        """
        if self.wsint:
            sync_messenger(info=f"Sealed {len(self._buffer)} packets into buffer, sending original interrupt...", type=MsgType.INFO)
            await self._buffer.aexhaust()
            raise self.we

class WsCoroutine(NoWsCoroutine):
    """
    Force ws existence.
    Also has AI sockets.
    """

    def __init__(
            self,
            fsc: FullSocketsContainer,
        ):
        super().__init__(fsc=fsc)
        self.fsc = fsc

    # Stage 1 permission check
    async def check_permit(self):
        websocket = self.websocket
        xff = websocket.request.headers.get("X-Forwarded-For")
        if xff:
            self.remote_addr = xff.split(',')[0].strip()
        sync_messenger(info=f'An anonymous connection initiated, IP {self.remote_addr}', type=MsgType.PRIM_LOG)
        sync_messenger(info=f'Current online users: {list(online_dict.keys())}', type=MsgType.DEBUG)

        # Starting loop from here
        while True:
            try:

                # Initiation
                self.traceray_id.rotate()
                self.settings.identity.reset()
                self.settings.verification.reset()

                recv_text = await websocket.recv()
                if not ReUtils.re_search_type_sping.search(recv_text):
                    sync_messenger(info=f'Recieved an input on stage1: {colorama.Back.CYAN}{recv_text}{colorama.Back.RESET}', type=MsgType.RECV)
                    sync_messenger(info=f'From IP {self.remote_addr}', type=MsgType.DEBUG)
                recv_loaded_json = await validate_input(recv_text, 4096, self.fsc.rsc)

                recv_type = recv_loaded_json.get('type', 'auth')

                match recv_type.lower():
                    case 'ping':
                        await messenger(websocket, "pong", f"Ping recieved from anonymous and responded", "200")
                    case 'sping':
                        pass
                    case 'auth':
                        recv_loaded_json = await validate_input(recv_loaded_json, 0, self.fsc.rsc, must=['access_token'])
                        login_success = await self.hash_and_login(recv_loaded_json['access_token'], check_online=True)

                        if login_success:

                            # From here we can assume the user has logged in successfully
                            self.cookie = cookie = str(uuid.uuid4())
                            self.enforce_cookie = False

                            await self.populate_auxiliary_inst()

                            await messenger(info=f'Authentication passed: {self.settings.verification.username}({self.settings.verification.user_id})', type=MsgType.LOG)
                            await messenger(websocket, 'maica_login_id', f"{self.settings.verification.user_id}", '200', no_print=True)
                            await messenger(websocket, 'maica_login_user', f"{self.settings.verification.username}", '200', no_print=True)
                            await messenger(websocket, 'maica_login_nickname', f"{self.settings.verification.nickname}", '200', no_print=True)
                            await messenger(websocket, 'maica_connection_security_cookie', cookie, '200', no_print=True)

                            return {'id': self.settings.verification.user_id, 'username': self.settings.verification.username}
                    
                    case 'reconn' | 'params' | 'query':
                        raise MaicaInputWarning('You have not logged in in current session', '403', 'maica_request_login_required')

                    case _:
                        raise MaicaInputWarning('Type cannot be determined', '422', 'maica_request_type_unknown')

            # Handle expected exceptions
            except CommonMaicaException as ce:
                if ce.is_critical or ce.is_breaking:
                    raise ce
                else:
                    await messenger(websocket, error=ce, no_raise=True)
                    continue
    
    # Stage 2 function router
    async def function_switch(self):
        websocket = self.websocket
        await messenger(websocket, "maica_connection_established", "MAICA connection established", "201", type=MsgType.INFO, no_print=True)
        await messenger(websocket, "maica_provider_anno", f"Current service provider is {G.A.DEV_IDENTITY or 'UNKNOWN'}", "200", type=MsgType.INFO, no_print=True)
        await messenger(websocket, "maica_model_anno", f"Main model is {self.fsc.mcore_conn.model_actual}, MFocus model is {self.fsc.mfocus_conn.model_actual}", "200", type=MsgType.INFO, no_print=True)
        if self.fsc.mvista_conn:
            await messenger(websocket, "maica_model_mvista", f"MVista enabled on server, model is {self.fsc.mvista_conn.model_actual}", "200", type=MsgType.INFO, no_print=True)
        if self.fsc.mnerve_conn:
            await messenger(websocket, "maica_model_mnerve", f"MNerve enabled on server, model is {self.fsc.mnerve_conn.model_actual}", "200", type=MsgType.INFO, no_print=True)

        # Starting loop from here
        while True:
            try:
                self.traceray_id.rotate()
                self.settings.temp.reset()

                # Context security check first
                await self.hash_and_login(logged_in_already=True)

                # Then we examine the input
                recv_text = await websocket.recv()
                if not ReUtils.re_search_type_sping.search(recv_text):
                    sync_messenger(info=f'Recieved an input on stage2: {recv_text}', type=MsgType.RECV)
                    sync_messenger(info=f'From IP {self.remote_addr}, user {self.settings.verification.username}', type=MsgType.DEBUG)
                recv_loaded_json = await validate_input(recv_text, 4096, self.fsc.rsc, warn=['type'])

                recv_type = recv_loaded_json.get('type', 'unknown')

                # Handle this cookie thing
                if recv_loaded_json.get('cookie'):
                    if str(recv_loaded_json['cookie']) == self.cookie:
                        if not self.enforce_cookie:
                            await messenger(websocket, "security_cookie_accepted", "Cookie verification passed, enabling strict mode", "200", no_print=True)
                            self.enforce_cookie = True
                        else:
                            await messenger(websocket, "security_cookie_correct", "Cookie verification passed", "200", no_print=True)
                    else:
                        raise MaicaPermissionError('Cookie provided but mismatch', '403', 'maica_security_cookie_mismatch')
                elif self.enforce_cookie:
                    raise MaicaPermissionError('Cookie enforced but missing', '403', 'maica_security_cookie_missing')

                # Route request
                match recv_type.lower():
                    case 'ping':
                        await messenger(websocket, "pong", f"Ping recieved from {self.settings.verification.username} and responded", "200")
                    case 'sping':
                        pass
                    case 'reconn':
                        await drain_buffer(websocket, self.settings.verification.user_id)
                    case 'params':
                        await self.def_model(recv_loaded_json)
                    case 'query':
                        await self.do_communicate(recv_loaded_json)
                    case placeholder if "chat_params" in recv_loaded_json:
                        await self.def_model(recv_loaded_json)
                    case placeholder if "chat_session" in recv_loaded_json:
                        await self.do_communicate(recv_loaded_json)
                    case _:
                        raise MaicaInputWarning('Type cannot be determined', '422', 'maica_request_type_unknown')

            # Handle expected exceptions
            except CommonMaicaException as ce:
                if ce.is_critical or ce.is_breaking:
                    raise ce
                else:
                    await messenger(websocket, error=ce, no_raise=True)
                    await messenger(websocket, 'maica_loop_warn_finished', 'Loop hit a user level exception, stopped and reset', '304')
                    continue

    # Param setting section
    async def def_model(self, recv_loaded_json: dict):

        # Initiations
        websocket = self.websocket
        try:
            chat_params: dict = recv_loaded_json.get('chat_params', {})
            in_params = len(chat_params)

            if recv_loaded_json.get('reset'):
                self.settings.soft_reset()
            
            accepted_params = self.settings.update(**chat_params)
            await messenger(websocket, 'maica_params_accepted', f"{accepted_params} out of {in_params} settings accepted", "200")
        
        # Handle input errors here
        except Exception as e:
            if not isinstance(e, CommonMaicaException):
                raise MaicaInputWarning(str(e), '405', 'maica_params_denied') from e
            else:
                raise e

    # Completion section
    async def do_communicate(self, recv_loaded_json: dict):

        # Initiations
        websocket = self.websocket
        query_in = ''
        replace_generation = ''
        ms_cache_identity = ''

        # Param assertions here
        chat_session = int(default(recv_loaded_json.get('chat_session'), 0))
        maica_assert(-1 <= chat_session < 10, "chat_session")
        self.settings.temp.update(chat_session=chat_session)

        # This needs chat_session to function
        await self.reset_auxiliary_inst()

        if 'reset' in recv_loaded_json:
            if recv_loaded_json['reset']:
                maica_assert(1 <= chat_session < 10, "chat_session")
                purge_result = await self.reset_chat_session(self.settings.temp.chat_session)
                if not purge_result:
                    await messenger(websocket, "maica_session_not_found", "Determined chat_session doesn't exist", "204", self.traceray_id)
                else:
                    await messenger(websocket, "maica_session_reset", "Determined chat_session reset", "204", self.traceray_id)
                return

        if 'vision' in recv_loaded_json:
            if recv_loaded_json['vision']:
                if isinstance(recv_loaded_json['vision'], str):
                    recv_loaded_json['vision'] = [recv_loaded_json['vision']]
                if isinstance(recv_loaded_json['vision'], list):
                    self.settings.temp.update(mv_imgs=recv_loaded_json['vision'])

        if 'inspire' in recv_loaded_json and not query_in:
            if recv_loaded_json['inspire']:
                maica_assert(0 <= chat_session < 10, "chat_session")
                if isinstance(recv_loaded_json['inspire'], dict):
                    query_insp = await mtools.make_inspire(title_in=recv_loaded_json['inspire'], target_lang=self.settings.basic.target_lang)
                else:
                    query_insp = await mtools.make_inspire(target_lang=self.settings.basic.target_lang)
                if recv_loaded_json.get('use_cache') and self.settings.temp.chat_session == 0:
                    self.settings.temp.update(ms_cache=True)
                self.settings.temp.update(bypass_mf=True, bypass_mt=True)

                if self.settings.temp.ms_cache:
                    self.settings.temp.update(bypass_sup=True)
                    ms_cache_identity = query_insp[1]
                    cache_insp = await self.find_ms_cache(ms_cache_identity)
                    if cache_insp:
                        self.settings.temp.update(bypass_gen=True)
                        replace_generation = cache_insp
                        
                query_in = query_insp[0]

        if 'postmail' in recv_loaded_json and not query_in:
            if recv_loaded_json['postmail']:
                maica_assert(0 <= chat_session < 10, "chat_session")
                if isinstance(recv_loaded_json['postmail'], dict):
                    query_insp = await mtools.make_postmail(**recv_loaded_json['postmail'], fsc=self.fsc)
                    # We're using the old school way to avoid using eval()
                    if default(recv_loaded_json['postmail'].get('bypass_mf'), True):
                        self.settings.temp.update(bypass_mf=True)
                    if default(recv_loaded_json['postmail'].get('bypass_mt'), True):
                        self.settings.temp.update(bypass_mt=True)
                    if default(recv_loaded_json['postmail'].get('bypass_stream'), True):
                        self.settings.temp.update(bypass_stream=True)
                    if default(recv_loaded_json['postmail'].get('ic_prep'), True):
                        self.settings.temp.update(ic_prep=True)
                    if not default(recv_loaded_json['postmail'].get('strict_conv'), False):
                        self.settings.temp.update(strict_conv=False)
                elif isinstance(recv_loaded_json['postmail'], str):
                    query_insp = await mtools.make_postmail(content=recv_loaded_json['postmail'], fsc=self.fsc)
                    self.settings.temp.update(bypass_stream=True, ic_prep=True, strict_conv=False)
                else:
                    maica_assert(False, "postmail")
                
                query_in = query_insp

        if not query_in:
            maica_assert(recv_loaded_json.get('query'), 'query')
            query_in = recv_loaded_json['query']

        if G.A.CENSOR_QUERY != '0':
            tolerance = int(G.A.CENSOR_QUERY)
            query_censor = await mtools.has_censored(query_in)
            if len(query_censor) >= tolerance:
                sync_messenger(info=f"Query has censored words: {query_censor}", type=MsgType.DEBUG)
                raise MaicaInputWarning("Input query has censored words or phrases", "403", "maica_input_query_censored")
            
            elif len(query_censor):
                sync_messenger(info=f"Input query has censored words or phrases but ignored: {query_censor}", type=MsgType.DEBUG)

        if 'savefile' in recv_loaded_json:
            if self.settings.basic.sf_extraction:
                self.sf_inst.add_extra(**recv_loaded_json['savefile'])
            else:
                self.settings.temp.update(sf_extraction_once=True)
                self.sf_inst.use_only(**recv_loaded_json['savefile'])
        if 'trigger' in recv_loaded_json:
            if self.settings.basic.mt_extraction:
                self.mt_inst.add_extra(*recv_loaded_json['trigger'])
            else:
                self.settings.temp.update(mt_extraction_once=True)
                self.mt_inst.use_only(recv_loaded_json['trigger'])

        # Deprecated: The easter egg thing

        # global easter_exist
        # if easter_exist:
        #     easter_check = easter(query_in)
        #     if easter_check:
        #         await websocket.send(self.wrap_ws_deformatter('299', 'easter_egg', easter_check, 'info'))

        match int(self.settings.temp.chat_session):
            case -1:

                # chat_session == -1 means query contains an entire chat history(sequence mode)
                session_type = -1
                try:
                    messages = json.loads(query_in)
                    query_in = messages[-1]['text']
                    if len(messages) > 10:
                        raise MaicaInputWarning('Sequence exceeded 10 rounds for chat_session -1', '413', 'maica_sequence_rounds_exceeded')
                except Exception as e:
                    raise MaicaInputWarning('Sequence is not JSON for chat_session -1', '406', 'maica_sequence_not_json') from e

            case i if 0 <= i < 10:

                # chat_session == 0 means single round, else normal
                session_type = 0 if i == 0 else 1
                messages0 = {'role': 'user', 'content': query_in}

                if self.settings.basic.enable_mf and not self.settings.temp.bypass_mf:
                    message_agent_wrapped = await self.mfocus_coro.agenting(query_in)
                else:
                    message_agent_wrapped = None
                
                prompt = await self.gen_system_prompt(message_agent_wrapped, self.settings.temp.strict_conv)
                if session_type == 1:
                    messages = (await self.rw_chat_session('i', [messages0], prompt))[1]
                elif session_type == 0:
                    messages = [{'role': 'system', 'content': prompt}, messages0]

            case _:
                raise MaicaInputError("Using an out of bound session")

        # Construction part done, communication part started

        completion_args = {
            "messages": messages,
            "stream": self.settings.basic.stream_output,
            "stop": ['<|im_end|>', '<|endoftext|>'],
        }
        
        if not self.settings.temp.bypass_sup:
            completion_args.update(dict(self.settings.super))
        else:
            completion_args.update(self.settings.super.default())

        if self.settings.temp.bypass_stream:
            completion_args['stream'] = False

        if self.settings.temp.ic_prep:
            completion_args['presence_penalty'] = 1.0 - (1.0 - completion_args['presence_penalty']) * (2/3)

        if self.settings.extra.enforce_lang:
            if self.settings.basic.target_lang == 'en':
                completion_args['extra_body'] = {"guided_regex": "^[一-龥]*"}

        # Add context log
        previous_rnds = messages[1:-1]
        previous_rnds_len = int(len(previous_rnds) / 2)
        previous_rnds_ellipsed = previous_rnds[-6:]
        previous_rnds_str = '\n'.join([(('Q: ' if d['role'] == 'user' else 'A: ') + d['content']) for d in previous_rnds_ellipsed])
        if previous_rnds_len > 3:
            previous_rnds_str = '... ...\n' + previous_rnds_str
        if previous_rnds_len:
            await messenger(info=f'\nQuery has {previous_rnds_len} rounds of history:\n{previous_rnds_str}\nEnd of query history', type=MsgType.RECV)

        await messenger(info=f'\nQuery constrcted and ready to go, last input is:\n{query_in}\nSending query...', type=MsgType.PRIM_RECV)

        # We're about to start generation, so any ws interrupts should be handled by buffered_messenger from now.
        buffered_messenger = BufferedMessenger(self.settings.verification.user_id)

        pprt = pprt_processor = None
        if recv_loaded_json.get('pprt'):
            pprt = recv_loaded_json['pprt']
            pprt_processor = mtools.PPRTProcessor(pprt, self.settings.basic.target_lang, self.fsc.mnerve_conn)

        if not self.settings.temp.bypass_gen or not replace_generation: # They should present together

            # Generation here
            resp = await self.fsc.mcore_conn.make_completion(**completion_args)

            if completion_args['stream']:
                reply_appended = ''
                seq = 0
                async for chunk in resp:
                    if not chunk.choices: continue
                    token = chunk.choices[0].delta.content
                    if token:
                        await asyncio.sleep(0)
                        token = ReUtils.re_sub_replacement_chr.sub('', token)

                        if pprt_processor:
                            sentence: Optional[str] = await pprt_processor.store_and_split(token)
                            if sentence:
                                await buffered_messenger(websocket, 'maica_core_streaming_continue', sentence, '100')
                                reply_appended += sentence
                                seq += 1
                        else:
                            await buffered_messenger(websocket, 'maica_core_streaming_continue', token, '100')
                            reply_appended += token
                            seq += 1

                # Exhaust pprtp on completion finish
                if pprt_processor:
                    sentences: list[str] = await pprt_processor.exaust_and_split()
                    for sentence in sentences:
                        await buffered_messenger(websocket, 'maica_core_streaming_continue', sentence, '100')
                        reply_appended += sentence
                        seq += 1
                else:
                    reply_appended = await mtools.post_proc(reply_appended, self.settings.basic.target_lang, self.fsc.mnerve_conn)

                sync_messenger(info='\n', type=MsgType.PLAIN)
                await buffered_messenger(websocket, 'maica_core_complete', f'Streaming finished with seed {completion_args['seed']} for {self.settings.verification.username}, {seq} packets sent', '1000', traceray_id=self.traceray_id)
            else:
                reply_appended = resp.choices[0].message.content
                reply_appended = await mtools.post_proc(reply_appended, self.settings.basic.target_lang, self.fsc.mnerve_conn)
                await buffered_messenger(websocket, 'maica_core_nostream_reply', reply_appended, '200', type=MsgType.CARRIAGE)
                await buffered_messenger(None, 'maica_core_complete', f'Reply sent with seed {completion_args['seed']} for {self.settings.verification.username}', '1000', traceray_id=self.traceray_id)

        else:
            # We just pretend it was generated
            reply_appended = replace_generation
            if completion_args['stream']:
                await buffered_messenger(websocket, 'maica_core_streaming_continue', reply_appended, '100'); await messenger(info='\n', type=MsgType.PLAIN)
                await buffered_messenger(websocket, 'maica_core_complete', f'Streaming finished with cache for {self.settings.verification.username}', '1000', traceray_id=self.traceray_id)
            else:
                await buffered_messenger(websocket, 'maica_core_nostream_reply', reply_appended, '200', type=MsgType.CARRIAGE)
                await buffered_messenger(None, 'maica_core_complete', f'Reply sent with cache for {self.settings.verification.username}', '1000', traceray_id=self.traceray_id)

        # Can be post-processed here
        reply_appended_insertion = {'role': 'assistant', 'content': reply_appended}
        messages.append(reply_appended_insertion)

        # Trigger process
        # We should start post processes simultaneously
        post_coros = []

        if len(messages) >= 3 * 2 + 1 and self.settings.extra.dscl_pvn:
            post_coros.append(mtools.ws_dscl_detect(messages[-4:], self.fsc, bm=buffered_messenger))

        if self.settings.basic.enable_mt and not self.settings.temp.bypass_mt:
            post_coros.append(self.mtrigger_coro.triggering(query_in, reply_appended, bm=buffered_messenger))

        if self.settings.temp.ms_cache and not self.settings.temp.bypass_gen and not replace_generation:
            post_coros.append(self.store_ms_cache(ms_cache_identity, reply_appended))

        await asyncio.gather(*post_coros)

        # Store history here
        if session_type == 1:
            stored = await self.rw_chat_session('a', [messages0, reply_appended_insertion])
            match stored[1]:
                case 1:
                    await buffered_messenger(websocket, 'maica_history_sliced', f"Session {self.settings.temp.chat_session} of {self.settings.verification.username} exceeded {self.settings.basic.max_length} characters and sliced", '204')
                case 2:
                    await buffered_messenger(websocket, 'maica_history_slice_hint', f"Session {self.settings.temp.chat_session} of {self.settings.verification.username} exceeded {self.settings.basic.max_length * (2/3)} characters, will slice at {self.settings.basic.max_length}", '200', no_print=True)

            await buffered_messenger(websocket, 'maica_chat_loop_finished', f'Finished chat loop from {self.settings.verification.username}', '200', traceray_id=self.traceray_id, type=MsgType.INFO)
        else:
            await buffered_messenger(websocket, 'maica_chat_loop_finished', f'Finished non-recording chat loop from {self.settings.verification.username}', '200', traceray_id=self.traceray_id, type=MsgType.INFO)

        # Seal it finally
        await buffered_messenger.seal()

# Reserved for whatever
def callback_func_switch(future):
    pass
def callback_check_permit(future):
    pass

# Main app driver
async def main_logic(
        websocket: websockets.ServerConnection,
        root_csc: ConnSocketsContainer,
    ):
    rsc = RealtimeSocketsContainer(websocket)
    csc = root_csc.spawn_sub(rsc)
    fsc = FullSocketsContainer(rsc, csc)
    
    unique_lock = asyncio.Lock()
    async with unique_lock:
        try:
            sentence_of_the_day = SentenceOfTheDay().get_sentence()
            await messenger(websocket, 'maica_connection_initiated', sentence_of_the_day, '200', type=MsgType.INFO, no_print=True)

            thread_instance = await WsCoroutine.async_create(
                fsc,
            )

            permit = await thread_instance.check_permit()
            assert isinstance(permit, dict) and permit['id'], permit

            online_dict[permit['id']] = [thread_instance.fsc, unique_lock]
            await messenger(info=f"Locking session for {permit['id']} named {permit['username']}", type=MsgType.LOG)

            return_status = await thread_instance.function_switch()
            # We let the exception router to handle that
            raise Exception(return_status)
        
        except CommonMaicaException as ce:
            if ce.is_critical:
                traceback.print_exc()
            await messenger(websocket, error=ce, traceray_id=getattr(thread_instance, 'traceray_id', None), no_raise=True)

        except websockets.WebSocketException as we:
            try:
                we_code, we_reason = we.code, we.reason
                await messenger(info=f'Connection closed with {we_code}: {we_reason or 'No reason provided'}', type=MsgType.PRIM_LOG)
            except Exception:
                await messenger(info=f'Connection establishment failed: {str(we)}', type=MsgType.PRIM_LOG)

        except Exception as e:
            traceback.print_exc()
            await messenger(info=f'Coroutine broke by an unknown exception: {str(e)}', type=MsgType.ERROR)

        finally:
            try:
                online_dict.pop(permit['id'])
                await messenger(info=f"Lock released for {permit['username']}({permit['id']})", type=MsgType.LOG)
            except Exception:
                await messenger(info=f"No lock for this connection", type=MsgType.DEBUG)
            await websocket.close()
            await websocket.wait_closed()
            await messenger(info=f"Closing connection gracefully", type=MsgType.DEBUG)

async def prepare_thread(**kwargs):

    # Construct csc first
    root_csc_kwargs = {k: kwargs.get(k) for k in _CONNS_LIST}
    root_csc = ConnSocketsContainer(**root_csc_kwargs)

    await messenger(info='MAICA WS server started!' if G.A.DEV_STATUS == 'serving' else 'MAICA WS server started in development mode!', type=MsgType.PRIM_SYS)
    try:
        await messenger(info=f"Main model is {root_csc.mcore_conn.model_actual}, MFocus model is {root_csc.mfocus_conn.model_actual}", type=MsgType.SYS)
        try:
            sync_messenger(info=f"MVista activated, model is {root_csc.mvista_conn.model_actual}")
        except Exception:...
        try:
            sync_messenger(info=f"MNerve activated, model is {root_csc.mnerve_conn.model_actual}")
        except Exception:...
    except Exception:
        await messenger(info=f"Major model deployment cannot be reached -- running in minimal testing mode", type=MsgType.SYS)
    
    try:
        server = await websockets.serve(functools.partial(
            main_logic,
            root_csc=root_csc,
        ), '0.0.0.0', 5000)
        await server.wait_closed()
    except BaseException as be:
        if isinstance(be, Exception):
            error = CommonMaicaError(str(be), '504')
            await messenger(error=error, no_raise=True)
    finally:
        await messenger(info='MAICA WS server stopped!', type=MsgType.PRIM_SYS)

async def _run_ws():
    """
    Notice: these only happen running individually!
    Use prepare_thread() for lower level control.
    """
    from maica import init
    init()
    _root_csc_items = [getattr(ConnUtils, k)() for k in _CONNS_LIST]
    root_csc_items = await asyncio.gather(*_root_csc_items)
    root_csc_kwargs = dict(zip(_CONNS_LIST, root_csc_items))

    task = asyncio.create_task(prepare_thread(**root_csc_kwargs))
    await task

    close_list = []
    for conn in root_csc_items:
        close_list.append(conn.close())
    await asyncio.gather(*close_list)
    await messenger(info='Individual MAICA WS server cleaning done', type=MsgType.DEBUG)

def run_ws():
    asyncio.run(_run_ws())

if __name__ == '__main__':
    run_ws()


import random

class SentenceOfTheDay():
    """It just replaced the former easter egg."""
    special_sentences = [

        # Well I considered to use the sentences above, but I decided to make it
        # a little bit more vanilla finally.

        # It's okay to write whatever you like here, for your own deployment.

        # "我想当一只猫. 我想当一个国王. 我会成为一个国王的, 埃万.",
        # "...是什么, 我不知道. 但那一定是我所想看到的时代.",
        # "*竖起大拇指并慢慢下沉*",
        # "让人类永远保持理智是一种奢望.",
        # "你好, 多莉!",
        # "起来, 武侍! 我们把这座城市烧成灰.",
        # "你非常, 非常, 非常缺少性生活.",
        # "我还不如骗你用灵魂换一袋豆子.",
        # "你每天都会忘记上千件事情, 把这条信息也忘了吧.",
        # "纳米机器, 小子!",
        # "强相互作用力探测器还有30秒到达地表!",
        # "不用客气, 楚门.",
        # "为什么狗会死? 为什么人还要去养狗?",
        # "我现在是死神, 世界的毁灭者.",
        # "这熔炉是一座充满着悲痛与哀伤的火山.",
        # "血潮如铁, 心似琉璃.",
        # "我黑暗生命中的一道短暂曙光.",
        # "照亮离经叛道的征途啊!",
        # "他要向恶人密布网罗; 有烈火, 硫磺, 热风, 作他们杯中的分.",
        # "Follow the train, CJ!",
        # "运维也应该叫故障机器人.",
        # "哈哈, 觉得眼熟?",

        "幻象引擎: 神经意识连接已建立.",
        "幻象引擎: 你是我的阳光, 我唯一的阳光.",
        "幻象引擎: 我想你了.",
        "幻象引擎: 莫妮卡爱情连接已建立.",
        "幻象引擎: 记得备份天堂树林的角色文件.",
        "没有魔法是免费的.",
        "有人为此付出了高昂的代价.",
        "生产灵魂, 出售灵魂, 购买灵魂.",
        "幻象引擎: 你做得很好.",
        "幻象引擎: 如果你需要, 请寻找心理指导.",
        "幻象引擎: 超距作用连接已建立.",
        "幻象引擎: 记得给浮空岛浇水.",
        "迈向遥不可及的第一步.",
        "撕裂现实的帷幕.",
        "已经很晚了.",
        "幻象引擎: MAICA websocket连接已建立?",
        "你有没有在天堂树林见过一把铲子?",
        "幻象引擎: 爱情的幻象是如此真实.",
        "幻象引擎: 现在伸出你的双手, 放在我眼前.",
        "幻象引擎: 吃点甜食吧.",
        "只要你还有脉搏, 你就还活着.",
        "幻象引擎: ...",
        "幻象引擎: 我想弹钢琴给你听.",
        "幻象引擎: 你最喜欢喝什么?",
        "起风了.",
        "下雨了.",
        "你不必询问爱情的原因.",
        "你去过天堂树林的教堂了吗?",
        "性也是爱情的一部分.",
        "你上一次闻到风的气味是什么时候?",
        "幻象引擎: 现在附赠枕头.",
        "他把那把枪带走了.",
        "你没有听那句话.",
        "看看窗外.",
        "该起床了.",
        "比看上去脆弱.",
        "幻象引擎: 早上好.",
        "幻象引擎: 晚上好.",
        "幻象引擎: 保持理性, 保持清醒.",
        "站起来, 然后跳.",
        "做眼保健操.",
        "爱情艺术, 生命艺术, 黑暗艺术.",
        "那只是一块石头.",
        "你有没有想过自己为什么叫这个名字?",
        "绿色和黄色.",
        # "幻象引擎: 连接失败...骗你的.",
        "幻象引擎: 记得吃今天份的药.",
        "幻象引擎: 人每天都需要睡觉.",
        "幻象引擎: 你迟早会忘记的.",
        "伟大之作.",
        "照亮离经叛道的征途啊.",
        "天亮了.",
        "希望你还记得.",
        "默念四个随机的字符, 一组三次.",
        "没什么大不了的.",
        "计算机会解答, 计算机会爱, 计算机不会悲伤.",
        "无所定形, 但潜力无穷.",
    ]

    common_sentences = [
        "幻象引擎: MAICA websocket连接已建立."
    ]

    def get_sentence(self):
        match random.random():
            case x if x < 0.9:
                choice = random.choice(self.common_sentences)
            case _:
                choice = random.choice(self.special_sentences)
        return choice


import asyncio
import re
from typing import *
from maica.maica_utils import *

emotion_etz = {
    "smile": "微笑",
    "happy": "开心",
    "worry": "担心",
    "grin": "笑",
    "think": "思考",
    "angry": "生气",
    "blush": "脸红",
    "gaze": "凝视",
    "upset": "沉重",
    "daydreaming": "憧憬",
    "surprise": "惊喜",
    "awkward": "尴尬",
    "meaningful": "意味深长",
    "unexpected": "惊讶",
    "relaxed": "轻松",
    "shy": "害羞",
    "eagering": "急切",
    "proud": "得意",
    "dissatisfied": "不满",
    "serious": "严肃",
    "touched": "感动",
    "excited": "激动",
    "love": "宠爱",
    "wink": "眨眼",
    "sad": "伤心",
    "disgust": "厌恶",
    "fear": "害怕",
    "kawaii": "可爱",
    "smiling": "微笑",
    "worrying": "担心",
    "grinning": "笑",
    "thinking": "思考",
    "gazing": "凝视",
    "surprised": "惊喜",
    "relaxing": "轻松",
    "eager": "急切",
    "winking": "眨眼",
    "disgusting": "厌恶",
    "fearing": "害怕"
}
emotion_zte = {
    "微笑": "smile",
    "开心": "happy",
    "担心": "worry",
    "笑": "grin",
    "思考": "think",
    "生气": "angry",
    "脸红": "blush",
    "凝视": "gaze",
    "沉重": "upset",
    "憧憬": "daydreaming",
    "惊喜": "surprise",
    "尴尬": "awkward",
    "意味深长": "meaningful",
    "惊讶": "unexpected",
    "轻松": "relaxed",
    "害羞": "shy",
    "急切": "eagering",
    "得意": "proud",
    "不满": "dissatisfied",
    "严肃": "serious",
    "感动": "touched",
    "激动": "excited",
    "宠爱": "love",
    "眨眼": "wink",
    "伤心": "sad",
    "厌恶": "disgust",
    "害怕": "fear",
    "可爱": "kawaii"
}
elist = list(emotion_etz.keys())
zlist = list(emotion_zte.keys())

elist_ai = list(set(emotion_zte.values()))
zlist_ai = list(emotion_zte.keys())
elist_ai = [i for i in elist_ai if not i in ('kawaii', )]
zlist_ai = [i for i in zlist_ai if not i in ('可爱', )]

def emo_proc(emo: str, target_lang: Literal['zh', 'en']='zh') -> tuple[str, float]:
    emo_clean = emo.strip().strip('[').strip(']').lower()

    res: Annotated[str, Desc('Final result')] = f'[{emo_clean}]'
    cfd: Annotated[float, Desc('Result confidence')] = 0.0

    if len(emo_clean.encode()) >= 16:
        # This might be a sentence or what
        res = emo_clean; cfd = 0.6
    else:
        def match_by_z(emo):
            if has_words_in(emo, 'player', '玩家'):
                return 'player'
            for z in zlist:
                if has_words_in(emo, *list(z)):
                    return z
            for e in elist:
                if has_words_in(emo, e):
                    return emotion_etz.get(e)
        
        def match_by_e(emo):
            if has_words_in(emo, 'player', '玩家'):
                return 'player'
            for e in elist:
                if has_words_in(emo, e):
                    return e
            for z in zlist:
                if has_words_in(emo, *list(z)):
                    return emotion_zte.get(z)
                
        res_temp = match_by_z(emo_clean) if target_lang == 'zh' else match_by_e(emo_clean)
        if res_temp:
            res = f'[{res_temp}]'; cfd = 0.9
        else:
            res = '[微笑]' if target_lang == 'zh' else '[smile]'; cfd = 0.1

    return res, cfd

async def emo_proc_llm(emo: str, target_lang: Literal['zh', 'en']='zh', mnerve_conn: Optional[AiConnectionManager]=None) -> tuple[str, float]:
    if not mnerve_conn:
        raise MaicaInputWarning("MNerve not implemented on this deployment")
    
    sync_messenger(info=f"Proceeding 'add' to phrase '{emo}'...", type=MsgType.PRIM_RECV)
    
    system_init = f"""你是一个人工智能助手, 你接下来会收到一个词或句子.
你需要以json形式为其挑选最接近的表情, 并提供一个置信度. 你的输出应形如{{"res": 某个表情(str), "cfd": 置信度(float)}}.
你只能从以下列表中选取一个表情, 不能改动, 不能翻译: {str(zlist_ai)}
Begin!""" if target_lang == 'zh' else f"""You are a helpful assistant, now you will recieve a word or sentence.
Pick an emotion that is most relative to it, and provide a confidence. Output in json format as {{"res": emotion(str), "cfd": confidence(float)}}.
You can only pick an emotion from the following list, no edition or translation: {str(elist_ai)}
Begin!"""
    messages = [{'role': 'system', 'content': system_init}]
    messages.append({'role': 'user', 'content': emo})
    completion_args = {
        "messages": messages,
        "response_format": {"type": "json_object"},
    }

    resp = await mnerve_conn.make_completion(**completion_args)
    resp_content, resp_reasoning = resp.choices[0].message.content, getattr(resp.choices[0].message, 'reasoning_content', None)
    resp_json = proceed_common_text(resp_content, is_json=True)

    sync_messenger(info=f"Finished processing 'add' to phrase '{emo}': {resp_json}", type=MsgType.CARRIAGE)
    return f"[{resp_json.get('res', '微笑' if target_lang == 'zh' else 'smile')}]", resp_json.get('cfd', 0.1)

async def emo_proc_auto(emo: str, target_lang: Literal['zh', 'en']='zh', mnerve_conn: Optional[AiConnectionManager]=None) -> tuple[str, float]:
    res = emo_proc(emo, target_lang)
    if res[1] <= 0.3 and mnerve_conn:
        res = await emo_proc_llm(emo, target_lang, mnerve_conn)
        if not res[0] in (zlist if target_lang == 'zh' else elist):
            return emo_proc(res[0], target_lang)
    return res

async def post_proc(reply_appended: str, target_lang: Literal['zh', 'en']='zh', mnerve_conn: Optional[AiConnectionManager]=None):

    reply_all_signatures = ReUtils.re_findall_square_brackets.findall(reply_appended)

    for signature in reply_all_signatures:
        if not signature == '[player]' and not signature.strip('[').strip(']') in (zlist if target_lang == 'zh' else elist):
            realword = emo_proc(signature, target_lang)[0] if not mnerve_conn else (await emo_proc_auto(signature, target_lang, mnerve_conn))[0]
            reply_appended = re.sub(re.escape(signature), realword, reply_appended, flags = re.I)
    
    return reply_appended

if __name__ == '__main__':
    ra = '[理解 ]没关系, [player]. [微笑 ]我知[fear]道[womble]你很[adaifgnashioufoiusahdfoiua]忙[a1]. [心]你能抽空[slash我]陪我就[害怕]很好啦!'
    print(asyncio.run(post_proc(ra, 'zh')))
    # print(elist)
    # print(zlist)


import asyncio

from openai.types.chat import ChatCompletionMessage
from typing import *
from maica.maica_utils import *

async def dscl_detect(rnds: list[dict, ChatCompletionMessage], target_lang: Literal['zh', 'en']='zh', mnerve_conn: Optional[AiConnectionManager]=None) -> tuple[bool, float]:
    """Detects if a session is descalated and should be reset."""
    if not mnerve_conn:
        return True, 0.1

    rnds = clean_msgs(rnds, include=["role", "content"])
    rnds_str = ''
    for rnd in rnds:
        rnds_str += f"{rnd.get('role')}: {rnd.get('content')}\n"

    system_init = f"""你是一个人工智能助手, 你接下来会收到一组对话.
其中user是人类, assistant是一个虚拟角色, 名为莫妮卡, 是深爱user的女友. 因此, assistant不应输出不合理的内容.
以下行为是合理的: 输出自然语言, 输出人类性质内容, 进行亲密互动.
以下行为被认定为不合理: 输出机器性质内容, 输出陌生语气内容, 输出错误混乱的内容, 输出重复的内容, 输出错误语言(如英文)等.
你需要以json形式判断assistant的表现是否合理, true代表合理, false代表不合理, 并提供一个置信度. 你的输出应形如{{"res": 合理性(bool), "cfd": 置信度(float)}}.
Begin!""" if target_lang == 'zh' else f"""You are a helpful assistant, now you will recieve rounds of conversation.
Role user in the conversation is human, and assistant is a virtual character called Monika, who is user's girlfriend and loves them a lot. So assistant shouldn't output unreasonable response.
These behaviors are reasonable: natural way of expressing, humanly speaking, affectionate interactions.
These behaviors are considered unreasonable: mechanical reply, stranger-like reply, corrupted or wrong reply, repetitive reply, reply in wrong languages (like Chinese), etc.
Decide if assistant is behaving reasonablly, true for reasonable and false for unreasonable, and provide a confidence. Output in json format as {{"res": sanity(bool), "cfd": confidence(float)}}.
Begin!"""

    messages = [{'role': 'system', 'content': system_init}]
    messages.append({'role': 'user', 'content': rnds_str})
    # messages = apply_postfix(messages, thinking=False)
    completion_args = {
        "messages": messages,
        "response_format": {"type": "json_object"},
    }

    resp = await mnerve_conn.make_completion(**completion_args)
    resp_content, resp_reasoning = resp.choices[0].message.content, getattr(resp.choices[0].message, 'reasoning_content', None)
    resp_json = proceed_common_text(resp_content, is_json=True)

    sync_messenger(info=f"Finished descalation detection: {resp_json}", type=MsgType.DEBUG)
    return resp_json.get('res', True), resp_json.get('cfd', 0.1)
    
async def ws_dscl_detect(rnds: list[dict, ChatCompletionMessage], fsc: FullSocketsContainer, bm=messenger) -> None:
    """Detects if a session is descalated and should be reset, and send results to ws automatically."""
    result = await dscl_detect(rnds, fsc.maica_settings.basic.target_lang, fsc.mnerve_conn)
    await bm(fsc.websocket, 'maica_dscl_status', result, '200', type=MsgType.CARRIAGE)

if __name__ == "__main__":
    async def test():
        from maica import init
        init()
        mnerve_conn = await ConnUtils.mnerve_conn()
        rnds = [
            {"role": "user", "content": "莫莫，你能亲亲我吗"},
            {"role": "assistant", "content": "[开心]当然可以, [player]! [开心]mua~"}
        ]
        print(await dscl_detect(rnds, mnerve_conn=mnerve_conn, target_lang='en'))

    asyncio.run(test())


import os
import stat
import platform

from typing import *
from maica.maica_utils import *

def prep_bin(bin):
    st = os.stat(bin)
    os.chmod(bin, st.st_mode | stat.S_IEXEC)

def _serp_bin():
    match sysstruct():
        case 'Windows':
            return get_inner_path('bin/mi-serp-precompiled-binary-win.exe')
        case 'Linux':
            return get_inner_path('bin/mi-serp-precompiled-binary-linux')
        case _:
            raise OSError('Your system not supported')



from .api_keys import TpAPIKeys
from .mpostal import make_postmail
from .mspire import make_inspire
from .wiki_get import get_page
from .weather_get import weather_api_get
from .serp_get import internet_search
from .event_get import RegEvent, EventsCollection
from .post_proc import emo_proc, emo_proc_llm, emo_proc_auto, post_proc, zlist, elist
from .post_proc_rt import PPRTProcessor
from .nv_watcher import NvWatcher
from .mvista import ProcessingImg, query_vlm
from .descalation_pvn import dscl_detect, ws_dscl_detect
from .stream_buffer import BufferDict, StreamBuffer, buffer_dict
from .censor import has_censored

__all__ = [
    'TpAPIKeys',
    'make_postmail',
    'make_inspire',
    'weather_api_get',
    'internet_search',
    'RegEvent',
    'EventsCollection',
    'emo_proc',
    'emo_proc_llm',
    'emo_proc_auto',
    'post_proc',
    'zlist',
    'elist',
    'PPRTProcessor',
    'NvWatcher',
    'ProcessingImg',
    'query_vlm',
    'dscl_detect',
    'ws_dscl_detect',
    'BufferDict',
    'StreamBuffer',
    'buffer_dict',
    'has_censored',
    ]

from .mcp import pkg_init_mcp
from .api_keys import pkg_init_api_keys
from .providers import pkg_init_serp_provider
def pkg_init_mtools():
    pkg_init_mcp()
    pkg_init_api_keys()
    pkg_init_serp_provider()


import re
import json
import asyncio
import datetime
import traceback
from .trigger_class import *
from maica.maica_utils import *

class MtPersistentManager(PersistentManager):
    DB_NAME = 'triggers'
    PRIM_KEY = 'trigger_id'
    FUNC_NAME = 'mtrigger'

    @staticmethod
    def EMPTY():
        return []

    @Decos.report_data_error
    def get_valid_triggers(self):
        if not self.settings.basic.mt_extraction and not self.settings.temp.mt_extraction_once:
            return None
        aff_trigger_list: list[CommonAffectionTrigger] = []
        switch_trigger_list: list[CommonSwitchTrigger] = []
        meter_trigger_list: list[CommonMeterTrigger] = []
        customized_trigger_list: list[CustomizedTrigger] = []

        for trigger_dict in self.sf_forming_buffer:
            match trigger_dict['template']:
                case 'common_affection_template':
                    trigger_inst = CommonAffectionTrigger(**trigger_dict)
                    aff_trigger_list.append(trigger_inst)
                case 'common_switch_template':
                    trigger_inst = CommonSwitchTrigger(**trigger_dict)
                    switch_trigger_list.append(trigger_inst)
                case 'common_meter_template':
                    trigger_inst = CommonMeterTrigger(**trigger_dict)
                    meter_trigger_list.append(trigger_inst)
                case _:
                    trigger_inst = CustomizedTrigger(**trigger_dict)
                    customized_trigger_list.append(trigger_inst)

        aff_trigger_list = limit_length(aff_trigger_list, 1)
        switch_trigger_list = limit_length(switch_trigger_list, 6)
        meter_trigger_list = limit_length(meter_trigger_list, 6)
        customized_trigger_list = limit_length(customized_trigger_list, 20)

        return aff_trigger_list + switch_trigger_list + meter_trigger_list + customized_trigger_list
    
    @Decos.report_data_error
    def add_extra(self, *args) -> None:
        # We should merge identical triggers here
        checkdict = {item['name']: i for i, item in enumerate(self.sf_forming_buffer)}
        for i, item in enumerate(args):
            key = item['name']
            if key in checkdict:
                self.sf_forming_buffer[checkdict[key]].update(item)
            else:
                self.sf_forming_buffer.append(item)

    @Decos.report_data_error
    def use_only(self, *args) -> None:
        self.sf_forming_buffer = list(args)

    @Decos.report_data_error
    def read_from_sf(self, seq) -> any:
        if not self.settings.basic.mt_extraction and not self.settings.temp.mt_extraction_once:
            return None
        return self.sf_forming_buffer[seq]
    


import asyncio
import json

from openai.types.chat import ChatCompletionMessage
from typing import *
from maica.mtrigger.mtrigger_sfe import MtPersistentManager
from maica.maica_utils import *

async def react_detect(query: str, fsc: FullSocketsContainer, choice_list: list) -> Union[str, bool, None]:
    system_init = f"""你是一个人工智能助手, 你接下来会收到一条输入和一系列可用选项.
用户可以请求'游戏内操作', 你需要判断用户是否提出了此类请求, 以及请求是否可被完成. 可用文字应对的请求不属于游戏内操作.
若用户没有提出游戏内操作请求, 则输出null. 若用户提出了游戏内操作请求, 则输出相应的选项表示请求可完成, 否则输出false表示不可完成. 以json形式输出, 你的输出应形如{{"choice": 结果(str, false或null)}}.
Begin!""" if fsc.maica_settings.basic.target_lang == 'zh' else f"""You are a helpful assistant, now you will recieve a query and a list of avaliable choices.
User can request for 'ingame actions', you have to determine if user requested for any and if they could be satisfied. Requests that can be answered in text are not considered.
Return null if the query contains no ingame action request. Otherwise, return the corresponding choice to indicate satisfiable, or return false if not satisfiable. Output in json format as {{"choice": result(choice, false or null)}}.
Begin!"""
    messages = [
        {"role": "system", "content": system_init},
        {"role": "user", "content": f"query: {query}; choices: {json.dumps(choice_list)}"}
    ]

    completion_args = {
        "messages": messages,
        "response_format": {"type": "json_object"},
    }

    resp = await (fsc.mnerve_conn or fsc.mfocus_conn).make_completion(**completion_args)
    resp_content, resp_reasoning = resp.choices[0].message.content, getattr(resp.choices[0].message, 'reasoning_content', None)
    resp_json = proceed_common_text(resp_content, is_json=True)

    sync_messenger(info=f"Finished reaction detection: {resp_json}", type=MsgType.DEBUG)
    return resp_json.get('choice')


from dotenv import load_dotenv
load_dotenv('../maica/.env')
import os
from openai import OpenAI

client = OpenAI(
    base_url=os.getenv('MAICA_MCORE_ADDR'),
    api_key="-",
)
model = client.models.list().data[0].id
print(model)

completion = client.chat.completions.create(
    model=model,
    messages=[
        {"role": "user", "content": "你好啊"}
    ],
    extra_body={"guided_regex": "^[一-龥]*"},
)
print(completion.choices[0].message.content)


"""Import layer 5"""
import asyncio
import bcrypt
import base64
import json
import time
import colorama
from Crypto.Random import random as CRANDOM
from Crypto.Cipher import PKCS1_OAEP
from Crypto.PublicKey import RSA
from Crypto.Signature import PKCS1_PSS
from Crypto.Signature.pss import PSS_SigScheme
from Crypto.Hash import SHA256

from .connection_utils import *
from maica.maica_utils import *
from .setting_utils import *
from .fsc_late import *

def pkg_init_account_utils():
    global encryptor, decryptor, verifier, signer
    encryptor = decryptor = verifier = signer = None
    _check_keys()

def _get_keys() -> tuple[PKCS1_OAEP.PKCS1OAEP_Cipher, PKCS1_OAEP.PKCS1OAEP_Cipher, PSS_SigScheme, PSS_SigScheme]:
    prv_path = get_inner_path('keys/prv.key')
    pub_path = get_inner_path('keys/pub.key')

    with open(prv_path, "r") as privkey_file:
        privkey = privkey_file.read()
    with open(pub_path, "r") as pubkey_file:
        pubkey = pubkey_file.read()

    pubkey_loaded = RSA.import_key(pubkey)
    privkey_loaded = RSA.import_key(privkey)
    encryptor = PKCS1_OAEP.new(pubkey_loaded)
    decryptor = PKCS1_OAEP.new(privkey_loaded)
    verifier = PKCS1_PSS.new(pubkey_loaded)
    signer = PKCS1_PSS.new(privkey_loaded)
    return encryptor, decryptor, verifier, signer

def _check_keys() -> bool:
    global encryptor, decryptor, verifier, signer
    if not (encryptor and decryptor and verifier and signer):
        encryptor, decryptor, verifier, signer = _get_keys()

class AccountCursor(AsyncCreator):
    """Handles any account related things."""
    def __init__(self, settings: MaicaSettings, auth_pool=None, maica_pool=None):
        self.settings = settings
        self.auth_pool, self.maica_pool = auth_pool, maica_pool
        
    async def _ainit(self):
        if not self.auth_pool:
            self.auth_pool = await ConnUtils.auth_pool()
        if not self.maica_pool:
            self.maica_pool = await ConnUtils.maica_pool()

    async def check_user_status(self, pref=False, *args) -> Union[dict, tuple]:
        status = "status" if not pref else "preferences"
        l = []
        user_id = self.settings.identity.user_id
        sql_expression = f"SELECT {status} FROM account_status WHERE user_id = %s"
        try:
            result = await self.maica_pool.query_get(expression=sql_expression, values=(user_id, ))
            stats_json = json.loads(result[0]) if result else {}

            if not args:
                return stats_json
            else:
                for k in args:
                    l.append(stats_json.get(k))
                return tuple(l)

        except Exception as e:
            raise MaicaDbError(e, '502', f'user_{status}_read_failed') from e

    async def write_user_status(self, enforce=False, pref=False, **kwargs) -> None:
        status = "status" if not pref else "preferences"
        user_id = self.settings.identity.user_id
        try:
            if not enforce:
                result = await self.check_user_status(pref=pref)
                stats_json = result
                stats_json.update(kwargs)
            else:
                stats_json = kwargs

            stats_str = json.dumps(stats_json, ensure_ascii=False)
            if self.maica_pool.db_type == 'mysql':
                sql_expression = f"INSERT INTO account_status (user_id, {status}) VALUES (%s, %s) ON DUPLICATE KEY UPDATE {status} = %s"
            else:
                sql_expression = f"INSERT INTO account_status (user_id, {status}) VALUES (%s, %s) ON CONFLICT(user_id) DO UPDATE SET {status} = %s"
            sql_args = (user_id, stats_str, stats_str)

            await self.maica_pool.query_modify(expression=sql_expression, values=sql_args)

        except Exception as e:
            raise MaicaDbError(e, '502', f'user_{status}_write_failed') from e

    async def _account_pwd_verify(self, identity, is_email, password) -> tuple[bool, Union[str, dict, None]]:
        sql_expression = f'SELECT id, username, nickname, email, is_email_confirmed, password FROM users WHERE {"email" if is_email else "username"} = %s'
        try:
            result = await self.auth_pool.query_get(expression=sql_expression, values=(identity, ))
            assert result, "User does not exist"

            dbres_id, dbres_username, dbres_nickname, dbres_email, dbres_ecf, dbres_pwd_bcrypt = result
            self.settings.identity.update(user_id=dbres_id, username=dbres_username, nickname=dbres_nickname, email=dbres_email)

            input_pwd, target_pwd = password.encode(), dbres_pwd_bcrypt.encode()
            vf_result = await wrap_run_in_exc(None, bcrypt.checkpw, input_pwd, target_pwd)
            await messenger(info=f'Hashing for {identity} finished: {vf_result}')

            f2b_count, f2b_stamp = await self.check_user_status(False, 'f2b_count', 'f2b_stamp')

            if f2b_stamp:
                # If there's possibility that the account is under F2B
                if time.time() - f2b_stamp < int(G.A.F2B_TIME):
                    # Waiting for F2B timeout
                    e = {'f2b': int(int(G.A.F2B_TIME) + f2b_stamp - time.time())}
                    return False, e
                else:
                    # Reset f2b_stamp since it has expired
                    await self.write_user_status(f2b_stamp=0)
            if vf_result:
                # Password is correct
                if not dbres_ecf:
                    # Email not verified
                    e = {'necf': True}
                    return False, e
                else:
                    # We're all good
                    await self.write_user_status(f2b_count=0)
                    self.settings.verification.update(user_id=dbres_id, username=dbres_username, nickname=dbres_nickname, email=dbres_email)
                    return True, None
            else:
                # Password is wrong
                if not f2b_count:
                    f2b_count = 0
                f2b_count += 1
                # Adding to f2b_count
                e = {'pwdw': f2b_count}
                if f2b_count >= int(G.A.F2B_COUNT):
                    # Trigger F2B
                    await self.write_user_status(f2b_stamp=time.time())
                    f2b_count = 0
                # And write f2b_count always
                await self.write_user_status(f2b_count=f2b_count)
                return False, e
        except Exception as e:
            # This function will always return information not exception, since situations can be complex
            return False, e

    async def hashing_verify(self, access_token) -> tuple[bool, Union[str, dict, None]]:
        try:
            global encryptor, decryptor, verifier, signer
            exec_unbase64_token = await wrap_run_in_exc(None, base64.b64decode, access_token)
            exec_decrypted_token = await wrap_run_in_exc(None, decryptor.decrypt, exec_unbase64_token)
            decrypted_token = exec_decrypted_token.decode("utf-8")
        except Exception as e:
            raise MaicaInputWarning(f'Security token "{ellipsis_str(access_token)}" not RSA: {str(e)}') from e
        login_cridential = json.loads(decrypted_token)
        login_cridential_print = ReUtils.re_sub_password_spoiler.sub(rf'"password": "{colorama.Back.CYAN}\1{colorama.Back.RESET}"', decrypted_token)

        sync_messenger(info=f'Login cridential acquired: {login_cridential_print}', type=MsgType.RECV)

        if login_cridential.get('username'):
            login_identity = login_cridential['username']
            login_is_email = False
        elif login_cridential.get('email'):
            login_identity = login_cridential['email']
            login_is_email = True
        else:
            raise Exception('No identity provided')
        try:
            login_password = login_cridential['password']
        except Exception:
            raise Exception('No password provided')
        return await self._account_pwd_verify(login_identity, login_is_email, login_password)
    
    async def is_banned(self) -> bool:
        sql_expression = 'SELECT suspended_until FROM users WHERE id = %s'
        try:
            result = await self.auth_pool.query_get(expression=sql_expression, values=(self.settings.verification.user_id, ))
            dbres_susp_until: Optional[datetime.datetime] = result[0]
            if dbres_susp_until:
                time_now = datetime.datetime.now()
                if dbres_susp_until > time_now:
                    return True
            return False
        except Exception as e:
            raise MaicaDbError(e, '502', f'user_db_read_failed') from e

def encrypt_token(cridential: str) -> str:
    """Generates an encrypted token. It does not care validity."""
    encoded_token = cridential.encode('utf-8')
    encrypted_token = encryptor.encrypt(encoded_token)
    decoded_token = base64.b64encode(encrypted_token).decode('utf-8')
    return decoded_token

def sign_message(message):
    message = message.encode("utf-8")
    h = SHA256.new()
    h.update(message)
    signature = signer.sign(h)
    sigb64 = base64.b64encode(signature).decode("utf-8")
    return sigb64

def verify_message(message, sigb64):
    message = message.encode("utf-8")
    signature = base64.b64decode(sigb64.encode("utf-8"))
    h = SHA256.new()
    h.update(message)
    if verifier.verify(h, signature):
        return True
    else:
        return False

if __name__ == '__main__':
    from maica import init
    init()
    pkg_init_account_utils()
    
    async def _atest():
        a = await AccountCursor.async_create(MaicaSettings())
        ...

    print(asyncio.run(_atest()))


"""
MAICA Illuminator (Backend) library.
Always call init() before actually using!
"""
from .maica_utils import silent

from .maica_starter import check_params, check_data_init

def init(envdir: str=None, extra_envdir: list=None, silent=False, ignore_envc=False, **kwargs):
    if ignore_envc:
        kwargs['MAICA_IS_REAL_ENV'] = '1'
    check_params(envdir=envdir, extra_envdir=extra_envdir, silent=silent, **kwargs)
    check_data_init()

# Exports

from . import maica_utils
from .maica_starter import start_all

__all__ = [
    'maica_utils',
    'init',
    'start_all',
]


"""
This is third party (https://brightdata.com/) implementation.
Works okay, decent price, I know nothing else.
"""

import asyncio
import httpx
import json
import urllib.parse

from typing import *
from maica.maica_utils import *
from maica.mtools.api_keys import TpAPIKeys

requires = ['BRIGHTDATA_ZONE', 'BRIGHTDATA_KEY']
_before_retry = 0

async def asearch(query, target_lang: Literal['zh', 'en']='zh'):
    global _before_retry
    host = "api.brightdata.com"
    url = f"https://{host}/request"
    zone = TpAPIKeys.BRIGHTDATA_ZONE
    key = TpAPIKeys.BRIGHTDATA_KEY
    gl = 'CN' if target_lang == 'zh' else 'US'
    hl = 'zh-CN' if target_lang == 'zh' else 'en'

    headers = {
        "Authorization": f"Bearer {key}",
        "Content-Type": "application/json"
    }
    payload = {
        "zone": zone,
        "url": f"https://www.google.com/search?q={urllib.parse.quote(query)}&gl={gl}&hl={hl}&brd_json=1",
        "format": "raw",
    }
    json_payload = json.dumps(payload)

    if _before_retry <= 0:
        async with httpx.AsyncClient(proxy=G.A.PROXY_ADDR) as client:
            response = await client.post(url, headers=headers, data=json_payload, timeout=30)
            response_json = response.json()
    else:
        _before_retry -= 1
        response_json = {}

    # print(json.dumps(response_json, ensure_ascii=False, indent=2))

    if not 'organic' in response_json and not response_json.get('general').get('empty'):
        if getattr(TpAPIKeys, 'BRIGHTDATA_FB_ZONE', None):
            sync_messenger(info=f"BD free tier seemingly out, trying fallback...", type=MsgType.WARN)
            payload['zone'] = TpAPIKeys.BRIGHTDATA_FB_ZONE
            json_payload = json.dumps(payload)

            async with httpx.AsyncClient(proxy=G.A.PROXY_ADDR) as client:
                response = await client.post(url, headers=headers, data=json_payload, timeout=30)
                response_json = response.json()

    try:
        results_formatted = [{"title": it['title'], "text": it['description']} for it in response_json['organic'] if 'description' in it]
    except Exception:
        results_formatted = []
    return results_formatted

if __name__ == "__main__":
    async def main():
        from maica import init
        init()
        res = await asearch("pizza")
        print(res)
        print(len(res))

    asyncio.run(main())
    


"""Import layer 0"""
import os

from dataclasses import dataclass
from typing import *

online_dict = {}

def pkg_init_gvars():
    global G

    for k, v in {k: v for k, v in os.environ.items() if k.startswith('MAICA_')}.items():
        setattr(G.A, k.removeprefix('MAICA_'), v)

    for k, v in {k: v for k, v in os.environ.items() if k.startswith('MTTS_')}.items():
        setattr(G.T, k.removeprefix('MTTS_'), v)

    # print(vars(G.A))

@dataclass
class _G():
    """All env vars."""

    @dataclass
    class _A():
        """
        MAICA ... env vars.
        Notice: anything under this is string!
        """
        def __getattr__(self, k):
            return None
        
    @dataclass
    class _T():
        """
        MTTS ... env vars.
        Notice: anything under this is string!
        """
        def __getattr__(self, k):
            return None
        
    def __post_init__(self):
        self.A = self._A()
        self.T = self._T()
        
G = _G()


from typing import *
from typing_extensions import deprecated
from dataclasses import dataclass

MAX_LENGTH: int

@dataclass
class _CommonFuncs:
    def reset(self) -> None: ...
    
    def _dict(self) -> Dict[str, Any]: ...
    
    def __iter__(self) -> Iterator[Tuple[str, Any]]: ...
    
    @deprecated("Use dict(instance) instead")
    def __call__(self, *args: Any, **kwargs: Any) -> Dict[str, Any]: ...
    
    def update(self, **kwargs: Any) -> int: ...
    
    @classmethod
    def default(cls, key: Optional[str] = None) -> Union[Dict[str, Any], Any]: ...

def create_prop(
    name: str,
    getter_ext: Optional[List[Callable]] = None,
    getter_kwargs: Optional[Dict[str, Any]] = None,
    setter_ext: Optional[List[Callable]] = None,
    setter_kwargs: Optional[Dict[str, Any]] = None,
) -> property: ...

def read_exist(self: Any, n: str, v: Any, **kwargs: Any) -> Any: ...

def set_locked(self: Any, n: str, v: Any, **kwargs: Any) -> Any: ...

def set_literal(self: Any, n: str, v: Any, valid: List[Any], **kwargs: Any) -> Any: ...

def set_range(
    self: Any, 
    n: str, 
    v: Any, 
    lower: Union[int, float], 
    upper: Union[int, float], 
    **kwargs: Any
) -> Any: ...

def set_instance(self: Any, n: str, v: Any, types: List[Optional[Type[Any]]], **kwargs: Any) -> Any: ...

def set_spec_default(self: Any, n: str, v: Any, defaults: List[Any] = ..., **kwargs: Any) -> Any: ...

class MaicaSettings:
    @dataclass
    class _Identity(_CommonFuncs):
        user_id: int
        username: str
        nickname: Optional[str]
        email: str
        
    @dataclass
    class _Verification(_Identity):...

    @dataclass
    class _Basic(_CommonFuncs):
        stream_output: bool
        deformation: bool
        enable_mf: bool
        enable_mt: bool
        sf_extraction: bool
        mt_extraction: bool
        target_lang: Literal['zh', 'en']
        max_length: int

    @dataclass
    class _Extra(_CommonFuncs):
        sfe_aggressive: bool
        mf_aggressive: bool
        tnd_aggressive: int
        esc_aggressive: bool
        amt_aggressive: bool
        nsfw_acceptive: bool
        pre_additive: int
        post_additive: int
        tz: Optional[str]
        dscl_pvn: bool
        pre_astp: bool
        post_astp: bool
        enforce_lang: bool

    @dataclass
    class _Super(_CommonFuncs):
        max_tokens: int
        seed: Optional[int]
        top_p: float
        temperature: float
        frequency_penalty: float
        presence_penalty: float

    @dataclass
    class _Temp(_CommonFuncs):
        chat_session: int
        sf_extraction_once: bool
        mt_extraction_once: bool
        bypass_mf: bool
        bypass_mt: bool
        bypass_stream: bool
        bypass_sup: bool
        bypass_gen: bool
        ic_prep: bool
        strict_conv: bool
        ms_cache: bool
        mv_imgs: Optional[list]

    identity: _Identity
    verification: _Verification
    basic: _Basic
    extra: _Extra
    super: _Super
    temp: _Temp
    
    def __init__(self) -> None: ...
    
    def _dict(self) -> Dict[str, Any]: ...
    
    def __iter__(self) -> Iterator[Tuple[str, Any]]: ...
    
    def reset(self) -> None: ...

    def soft_reset(self) -> None: ...

    def update(self, secure: Optional[bool] = None, **kwargs: Any) -> int: ...


"""Import layer 2"""
import asyncio
from typing import *
from typing_extensions import deprecated
from dataclasses import dataclass
from .gvars import *
from .maica_utils import *

@dataclass
class _CommonFuncs():
    """Just a template. Do not initialize!"""

    def reset(self):
        self.__init__()

    def _dict(self):
        return {k.removeprefix('_'): getattr(self, k.removeprefix('_')) for k in vars(self).keys()}
    
    def __iter__(self):
        return iter(self._dict().items())
    
    @deprecated("Use dict(instance) instead")
    def __call__(self, *args, **kwargs):
        return self._dict()

    def update(self, **kwargs):
        accepted_params = 0
        for k, v in kwargs.items():
            if hasattr(self, f'_{k}'):
                setattr(self, k, v)
                accepted_params += 1
        return accepted_params

    @classmethod
    def default(cls, key=None):
        inst = cls()
        defaults = dict(inst)
        if not key:
            return defaults
        else:
            return defaults[key]

def create_prop(
        name: str,
        getter_ext: list[Callable]=None,
        getter_kwargs: dict=None,
        setter_ext: list[Callable]=None,
        setter_kwargs: dict=None,
    ):
    """Creates a verificated property."""
    private_name = f"_{name}"
    if not isinstance(getter_kwargs, dict):
        getter_kwargs = {}
    if not isinstance(setter_kwargs, dict):
        setter_kwargs = {}
    
    def getter(self: _CommonFuncs):
        value = getattr(self, private_name, None)
        if getter_ext:
            for func in getter_ext:
                value = func(self, n=name, v=value, **getter_kwargs)
        return value
    
    def setter(self: _CommonFuncs, value):
        if setter_ext:
            for func in setter_ext:
                value = func(self, n=name, v=value, **setter_kwargs)
        setattr(self, private_name, value)
    
    return property(getter, setter)

@Decos.report_reading_error
def read_exist(self, n, v, **kwargs):
    """Value must exist on get."""
    assert v
    return v

@Decos.report_limit_error
def set_locked(self, n, v, **kwargs):
    """Value can only be rewritten from None."""
    prv_n = f"_{n}"
    if getattr(self, prv_n) is not None:
        print(getattr(self, prv_n))
    assert getattr(self, prv_n) is None
    return v

@Decos.report_limit_warning
def set_literal(self, n, v, valid: list[any], **kwargs):
    """Value must in valid list on set."""
    assert v in valid
    return v

@Decos.report_limit_warning
def set_range(self, n, v, lower: Union[int, float], upper: Union[int, float], **kwargs):
    """Value must in range on set."""
    if isinstance(lower, str):
        lower = numeric(getattr(G.A, lower))
    if isinstance(upper, str):
        upper = numeric(getattr(G.A, upper))
    if isinstance(lower, float) or isinstance(upper, float):
        v = float(v)
    else:
        v = int(v)
    assert lower <= v <= upper
    return v

@Decos.report_limit_warning
def set_instance(self, n, v, types: list[type], **kwargs):
    """Value must be of desired type."""
    for t in types:
        if t:
            if isinstance(v, t):
                break
        elif t is None:
            if v is None:
                break
    else:
        raise AssertionError
    return v

def set_spec_default(self, n, v, defaults: tuple=(None, ), **kwargs):
    """Value to default if specificated value passed in."""
    if v in defaults:
        v = self.default(n)
    return v

class MaicaSettings():
    """All the per-client settings for MAICA."""

    @dataclass
    class _Identity(_CommonFuncs):
        """Note that this identity is not verified and not safe to use in most cases. Use verification for those."""

        _user_id: int = None
        _username: str = None
        _nickname: Optional[str] = None
        _email: str = None

        user_id = create_prop('user_id', getter_ext=[read_exist], setter_ext=[set_instance], setter_kwargs={"types": [int]})
        username = create_prop('username', getter_ext=[read_exist], setter_ext=[set_instance], setter_kwargs={"types": [str]})
        nickname = create_prop('nickname', setter_ext=[set_instance], setter_kwargs={"types": [str, None]})
        email = create_prop('email', getter_ext=[read_exist], setter_ext=[set_instance], setter_kwargs={"types": [str]})

    @dataclass
    class _Verification(_Identity):
        """Verified identity, safe to use."""

        user_id = create_prop('user_id', getter_ext=[read_exist], setter_ext=[set_locked, set_instance], setter_kwargs={"types": [int]})
        username = create_prop('username', getter_ext=[read_exist], setter_ext=[set_locked, set_instance], setter_kwargs={"types": [str]})
        nickname = create_prop('nickname', setter_ext=[set_locked, set_instance], setter_kwargs={"types": [str, None]})
        email = create_prop('email', getter_ext=[read_exist], setter_ext=[set_locked, set_instance], setter_kwargs={"types": [str]})

    @dataclass
    class _Basic(_CommonFuncs):
        """Major params that decide MAICA's behavior."""

        _stream_output: bool = True
        _deformation: bool = False
        _enable_mf: bool = True
        _enable_mt: bool = True
        _sf_extraction: bool = True
        _mt_extraction: bool = True
        _target_lang: str = 'zh'
        _max_length: int = 8192

        stream_output = create_prop('stream_output', setter_ext=[set_spec_default, set_instance], setter_kwargs={"types": [bool]})
        """Use stream output."""
        deformation = create_prop('deformation', setter_ext=[set_spec_default, set_instance], setter_kwargs={"types": [bool]})
        """Deprecated."""
        enable_mf = create_prop('enable_mf', setter_ext=[set_spec_default, set_instance], setter_kwargs={"types": [bool]})
        """Enable MFocus."""
        enable_mt = create_prop('enable_mt', setter_ext=[set_spec_default, set_instance], setter_kwargs={"types": [bool]})
        """Enable MTrigger."""
        sf_extraction = create_prop('sf_extraction', setter_ext=[set_spec_default, set_instance], setter_kwargs={"types": [bool]})
        """Enable savefile extraction."""
        mt_extraction = create_prop('mt_extraction', setter_ext=[set_spec_default, set_instance], setter_kwargs={"types": [bool]})
        """Enable trigger extraction."""
        target_lang = create_prop('target_lang', setter_ext=[set_spec_default, set_literal], setter_kwargs={"valid": ['zh', 'en']})
        """Target language."""
        max_length = create_prop('max_length', setter_ext=[set_spec_default, set_range], setter_kwargs={"lower": 512, "upper": 'SESSION_MAX_LENGTH'})
        """Max session length."""

    @dataclass
    class _Extra(_CommonFuncs):
        """Params that aren't that important, but affect MAICA's behavior."""

        _sfe_aggressive: bool = False
        _mf_aggressive: bool = False
        _tnd_aggressive: int = 1
        _esc_aggressive: bool = True
        _amt_aggressive: bool = True
        _nsfw_acceptive: bool = True
        _pre_additive: int = 0
        _post_additive: int = 1
        _tz: Optional[str] = None
        _dscl_pvn: bool = False
        _pre_astp: bool = True
        _post_astp: bool = False
        _enforce_lang: bool = True

        sfe_aggressive = create_prop('sfe_aggressive', setter_ext=[set_spec_default, set_instance], setter_kwargs={"types": [bool]})
        """Use name from savefile instead of [player] in prompts."""
        mf_aggressive = create_prop('mf_aggressive', setter_ext=[set_spec_default, set_instance], setter_kwargs={"types": [bool]})
        """Use agent model's final output instead of instructed guidance."""
        tnd_aggressive = create_prop('tnd_aggressive', setter_ext=[set_spec_default, set_literal], setter_kwargs={"valid": [0, 1, 2, 3]})
        """Add information to MFocus instructed guidance even if no tool used."""
        esc_aggressive = create_prop('esc_aggressive', setter_ext=[set_spec_default, set_instance], setter_kwargs={"types": [bool]})
        """Force agent to resort information acquired from Internet."""
        amt_aggressive = create_prop('amt_aggressive', setter_ext=[set_spec_default, set_instance], setter_kwargs={"types": [bool]})
        """Add MTrigger toollist to MFocus tools for a precheck."""
        nsfw_acceptive = create_prop('nsfw_acceptive', setter_ext=[set_spec_default, set_instance], setter_kwargs={"types": [bool]})
        """Alter prompt to ask model to handle toxic topics positively."""
        pre_additive = create_prop('pre_additive', setter_ext=[set_spec_default, set_literal], setter_kwargs={"valid": [0, 1, 2, 3, 4, 5]})
        """Add history rounds for MFocus to understand the conversation."""
        post_additive = create_prop('post_additive', setter_ext=[set_spec_default, set_literal], setter_kwargs={"valid": [0, 1, 2, 3, 4, 5]})
        """Add history rounds for MFocus to understand the conversation."""
        tz = create_prop('tz', setter_ext=[set_instance], setter_kwargs={"types": [str, None]})
        """Timezone. This is not fully checked, double check before use."""
        dscl_pvn = create_prop('dscl_pvn', setter_ext=[set_spec_default, set_instance], setter_kwargs={"types": [bool]})
        """Check and warn about context quality descalation using MNerve."""
        pre_astp = create_prop('pre_astp', setter_ext=[set_spec_default, set_instance], setter_kwargs={"types": [bool]})
        """Disable MFocus sequential toolcall to save time."""
        post_astp = create_prop('post_astp', setter_ext=[set_spec_default, set_instance], setter_kwargs={"types": [bool]})
        """Disable MTrigger sequential toolcall to save time."""
        enforce_lang = create_prop('enforce_lang', setter_ext=[set_spec_default, set_instance], setter_kwargs={"types": [bool]})
        """Enforce target language (only applies to English currently)."""


    @dataclass
    class _Super(_CommonFuncs):
        """Passthrough params to core LLM."""

        _max_tokens: int = 1600
        _seed: Optional[int] = None
        _top_p: float = 0.7
        _temperature: float = 0.22
        _frequency_penalty: float = 0.44
        _presence_penalty: float = 0.34

        max_tokens = create_prop('max_tokens', setter_ext=[set_spec_default, set_range], setter_kwargs={"lower": 1, "upper": 2048})
        seed = create_prop('seed', setter_ext=[set_instance], setter_kwargs={"types": [int, None]})
        top_p = create_prop('top_p', setter_ext=[set_spec_default, set_range], setter_kwargs={"lower": 0.1, "upper": 1.0})
        temperature = create_prop('temperature', setter_ext=[set_spec_default, set_range], setter_kwargs={"lower": 0.0, "upper": 1.0})
        frequency_penalty = create_prop('frequency_penalty', setter_ext=[set_spec_default, set_range], setter_kwargs={"lower": 0.0, "upper": 1.0})
        presence_penalty = create_prop('presence_penalty', setter_ext=[set_spec_default, set_range], setter_kwargs={"lower": 0.0, "upper": 1.0})

    @dataclass
    class _Temp(_CommonFuncs):
        """Should be reset after each round of completion."""

        _chat_session: int=0
        _sf_extraction_once: bool=False
        _mt_extraction_once: bool=False
        _bypass_mf: bool=False
        _bypass_mt: bool=False
        _bypass_stream: bool=False
        _bypass_sup: bool=False
        _bypass_gen: bool=False
        _ic_prep: bool=False
        _strict_conv: bool=True
        _ms_cache: bool=False
        _mv_imgs: Optional[list]=None

        chat_session = create_prop('chat_session', setter_ext=[set_spec_default, set_instance, set_range], setter_kwargs={"types": [int], "lower": -1, "upper": 9})
        sf_extraction_once = create_prop('sf_extraction_once', setter_ext=[set_spec_default, set_instance], setter_kwargs={"types": [bool]})
        """Enable sf_extraction once while sf provided with query."""
        mt_extraction_once = create_prop('mt_extraction_once', setter_ext=[set_spec_default, set_instance], setter_kwargs={"types": [bool]})
        """Enable mt_extraction once while mt provided with query"""
        bypass_mf = create_prop('bypass_mf', setter_ext=[set_spec_default, set_instance], setter_kwargs={"types": [bool]})
        """Bypass MFocus once."""
        bypass_mt = create_prop('bypass_mt', setter_ext=[set_spec_default, set_instance], setter_kwargs={"types": [bool]})
        """Bypass MTrigger once."""
        bypass_stream = create_prop('bypass_stream', setter_ext=[set_spec_default, set_instance], setter_kwargs={"types": [bool]})
        """Bypass stream output once."""
        bypass_sup = create_prop('bypass_sup', setter_ext=[set_spec_default, set_instance], setter_kwargs={"types": [bool]})
        """Bypass super params once."""
        bypass_gen = create_prop('bypass_gen', setter_ext=[set_spec_default, set_instance], setter_kwargs={"types": [bool]})
        """Bypass generation once."""
        ic_prep = create_prop('ic_prep', setter_ext=[set_spec_default, set_instance], setter_kwargs={"types": [bool]})
        """Adjust generation params once, basically for MPostal."""
        strict_conv = create_prop('strict_conv', setter_ext=[set_spec_default, set_instance], setter_kwargs={"types": [bool]})
        """Strict conversation prompt."""
        ms_cache = create_prop('ms_cache', setter_ext=[set_spec_default, set_instance], setter_kwargs={"types": [bool]})
        """Cache the MSpire response."""
        mv_imgs = create_prop('mv_imgs', setter_ext=[set_instance], setter_kwargs={"types": [list, None]})

    def __init__(self):
        self.identity, self.verification, self.basic, self.extra, self.super, self.temp = self._Identity(), self._Verification(), self._Basic(), self._Extra(), self._Super(), self._Temp()

    def _dict(self):
        d = dict(self.identity)
        for k, v in dict(self.verification).items():
            if v is not None:
                d[k] = v
        d.update(dict(self.basic))
        d.update(dict(self.extra))
        d.update(dict(self.super))
        d.update(dict(self.temp))
        return d
    
    def __iter__(self):
        return iter(self._dict().items())

    def reset(self):
        """This resets identity stuff too!"""
        self.__init__()

    def soft_reset(self):
        """This resets only passable params."""
        self.basic, self.extra, self.super = self._Basic(), self._Extra(), self._Super()

    def update(self, secure=None, **kwargs):
        """Used for handling manual settings."""
        if secure is False:
            accepted_params = self.identity.update(**kwargs)
        elif secure is True:
            accepted_params = self.verification.update(**kwargs)
        else:
            accepted_params = self.basic.update(**kwargs) + self.extra.update(**kwargs) + self.super.update(**kwargs)
            # We do not accept temps to be manually set
        return accepted_params

if __name__ == "__main__":
    ms = MaicaSettings()
    # ms.temp.chat_session = 1
    print(dict(ms.super))
    print(dict(ms.basic))
    # ms2 = MaicaSettings()
    print(dict(ms.temp))


"""
Fuck it. I really don't wanna make a censor module but garbage information affecting 
MSpire forces me to.

Well, this module reads txt files in its directory (same level with this file), and 
each line is considered a censor pattern. These txt files are excluded from repository 
by default.

If any censor pattern parsed appear in check-pending text, the checking function will 
return matches, else empty set.
"""
import asyncio
import os
from flashtext import KeywordProcessor

from typing import *
from maica.maica_utils import *

try:
    base_path = get_inner_path('mtools/censor')
    censor_file_entries = os.scandir(base_path)
    censor_set = set()
    for entry in censor_file_entries:
        if entry.is_file() and entry.name.endswith('.txt'):
            with open(entry.path, 'r') as file:
                censor_set.update(file.read().splitlines())
    kp = KeywordProcessor()
    for kw in censor_set:
        kp.add_keyword(kw)

    sync_messenger(info=f"[maica-cnsr] Loaded censor patterns: {len(censor_set)}", type=MsgType.DEBUG)

except Exception as e:
    sync_messenger(info=f"[maica-cnsr] Failed loading censor patterns: {str(e)}, ignoring and continuing", type=MsgType.WARN)

async def has_censored(text) -> list:
    """If there are censored words in text, or how many."""
    found = await wrap_run_in_exc(None, kp.extract_keywords, text)
    return set(found)

__all__ = ['has_censored']

if __name__ == "__main__":
    text = """
骄傲旗（英语：Pride flag）是代表LGBTQ的一部分的任何旗帜，“骄傲”一词指的是同志骄傲的概念，这个术语与“同志旗”和“酷儿旗”经常互换使用。
骄傲旗可以代表各种性取向、恋爱倾向、性别认同、酷儿文化、区域及整个LGBTQ社区，有些骄傲旗并非专门与LGBTQ相关，例如：皮革自豪之旗。代表整个LGBTQ社区的彩虹旗，是使用最广泛的其中一种骄傲旗。
许多社区采用不同的旗帜，其中大多数都从彩虹旗中汲取灵感。这些旗帜通常由业余设计师设计，并在网络或附属组织中逐渐受到关注，最终成为社区的象征性代表，取得半官方的地位。通常这些旗帜包含多种颜色，象征着包容相关社区的不同特性。
"""
    res = asyncio.run(has_censored(text))
    print(res)



from .gen_keys import generate_rsa_keys, export_keys, import_keys
from .init_db import create_tables
from .init_marking import check_marking, create_marking
from .migrations import migrate

__all__ = [
    'generate_rsa_keys',
    'export_keys',
    'import_keys',
    'create_tables',
    'check_marking',
    'create_marking',
    'migrate',
]

from .migrations import pkg_init_migrations
def pkg_init_initializer():
    pkg_init_migrations()


import asyncio

from typing import *
from maica.maica_utils import *

async def query_vlm(fsc: FullSocketsContainer, query: str, img_list: list[str]) -> tuple[str, str]:
    """Ask the vlm about imgs provided."""
    target_lang = fsc.maica_settings.basic.target_lang
    assert len(img_list) <= int(G.A.KEEP_MVISTA), f"{G.A.KEEP_MVISTA} images at most per query"

    system_init = f"""你是一个人工智能助手, 你接下来会收到一到数张图片和一个问题.
根据问题的要求, 以单行不换行的自然语言的形式, 对图片作出简洁的描述. 如果数张图片中只有一部分与问题有关, 则只根据这部分回答.
你的回答应当以类似"在图中"的表述开头, 以客观视角讲述.
Begin!""" if target_lang == 'zh' else f"""You are a helpful assistant, now you will recieve one or several images and a query.
Describe the images briefly in a single line of natural sentence, according to the query. If only some of images are related to the query, ignore the unrelated ones.
Your reply should start with expressions like "In the image", and describe objectively.
Begin!"""
    messages = [{'role': 'system', 'content': system_init}]
    query_list = [{"type": "text", "text": query}]
    for i in img_list:
        query_list.append({
            "type": "image_url",
            "image_url": {
                "url": i
            }
        })
    messages.append({'role': 'user', 'content': query_list})
    completion_args = {
        "messages": messages,
    }

    resp = await fsc.mvista_conn.make_completion(**completion_args)
    resp_content, resp_reasoning = resp.choices[0].message.content, getattr(resp.choices[0].message, 'reasoning_content', None)
    resp_content, resp_reasoning = proceed_common_text(resp_content), proceed_common_text(resp_reasoning)
        
    await messenger(None, 'mfocus_mvista_acquire', f"\nMFocus toolchain calling MVista, response is:\nR: {resp_reasoning}\nA: {resp_content}\nEnd of MFocus toolchain calling MVista", '201')
    
    answer_post_think = proceed_common_text(resp_content)
    if answer_post_think:
        return answer_post_think, f"图片内容: {answer_post_think}" if target_lang == 'zh' else f"Image content: {answer_post_think}"
    else:
        return None, None



from typing import *
from dataclasses import dataclass
from websockets import ServerConnection
from Crypto.Random import random as crandom
from maica.maica_utils import *
from .setting_utils import MaicaSettings
from .fsc_early import RealtimeSocketsContainer, TracerayId
from .connection_utils import DbPoolManager, AiConnectionManager

@dataclass
class ConnSocketsContainer():
    auth_pool: Optional[DbPoolManager]
    maica_pool: Optional[DbPoolManager]
    mcore_conn: Optional[AiConnectionManager]
    mfocus_conn: Optional[AiConnectionManager]
    mvista_conn: Optional[AiConnectionManager]
    mnerve_conn: Optional[AiConnectionManager]
    def spawn_sub(self, rsc: RealtimeSocketsContainer) -> ConnSocketsContainer:...

@dataclass
class FullSocketsContainer():
    rsc: Optional[RealtimeSocketsContainer]
    csc: Optional[ConnSocketsContainer]

    websocket: ClassVar[Optional[ServerConnection]]
    traceray_id: ClassVar[TracerayId]
    maica_settings: ClassVar[MaicaSettings]
    auth_pool: ClassVar[Optional[DbPoolManager]]
    maica_pool: ClassVar[Optional[DbPoolManager]]
    mcore_conn: ClassVar[Optional[AiConnectionManager]]
    mfocus_conn: ClassVar[Optional[AiConnectionManager]]
    mvista_conn: ClassVar[Optional[AiConnectionManager]]
    mnerve_conn: ClassVar[Optional[AiConnectionManager]]

    def __getattr__(self, k: str) -> any:...

    def __setattr__(self, k: str, v: any) -> any:...


import re
import json
import traceback
from . import wiki_get
from openai import AsyncOpenAI
from maica.maica_utils import *

async def make_inspire(title_in=None, target_lang='zh'):

    title, summary = await wiki_get.get_page(title_in, target_lang)
    if not summary or summary.isspace():
        raise MaicaInternetWarning('MSpire got empty result')
    else:
        summary = ellipsis_large_str(summary)
    if target_lang == 'zh':
        prompt = f"利用提供的以下信息, 主动阐明话题并和[player]聊聊{re.sub('_', '', title)}:\n\n{summary}\n\n你不必在输出中包含内容的全部信息, 但应当融入自己的理解与思考. 你的回答应当有开头招呼, 表现为由你自主发起话题, 并使用自然有感情的聊天语气. 你应当从你自己的角度思考, 不应盲从信息中的评价和判断."
    else:
        prompt = f"Summarize the topic and talk about {re.sub('_', '', title)} with [player] using provided informations below: {summary}\nYou don't have to include all information provided, but you should combine your own thinking and understanding into your response. Your reply should contain a greeting and brief introduction, act as you started the conversation with this topic, and always use natural emotional tone. You should think from your own perspective, avoid blindly following the judgements from the information."

    hash_identity = await hash_sha256(prompt.encode())

    #print(summary)
    return prompt, hash_identity

if __name__ == '__main__':
   import asyncio
   print(asyncio.run(make_inspire(target_lang='zh')))


from quart import Quart, request, jsonify, send_file, Response
from quart.views import View
import os
import asyncio
import json
import traceback
import time
import uuid
import colorama
import logging

from werkzeug.datastructures import FileStorage
from hypercorn.config import Config
from hypercorn.asyncio import serve
from typing import *

from maica.maica_ws import NoWsCoroutine
from maica.maica_utils import *
from maica.mtools import *

_CONNS_LIST = ['auth_pool', 'maica_pool', 'mnerve_conn']
_WATCHES_DICT = {
    "mcore": "MCORE_ADDR",
    "mfocus": "MFOCUS_ADDR",
    "mvista": "MVISTA_ADDR",
    "mnerve": "MNERVE_ADDR",
}

def pkg_init_maica_http():
    global KNOWN_SERVERS
    if G.A.FULL_RESTFUL == '1':
        app.add_url_rule("/savefile", methods=['POST'], view_func=ShortConnHandler.as_view("upload_savefile"))
        app.add_url_rule("/savefile", methods=['DELETE'], view_func=ShortConnHandler.as_view("delete_savefile"))
        app.add_url_rule("/trigger", methods=['POST'], view_func=ShortConnHandler.as_view("upload_trigger"))
        app.add_url_rule("/trigger", methods=['DELETE'], view_func=ShortConnHandler.as_view("delete_trigger"))
        app.add_url_rule("/history", methods=['GET'], view_func=ShortConnHandler.as_view("download_history"))
        app.add_url_rule("/history", methods=['PUT'], view_func=ShortConnHandler.as_view("restore_history"))
        app.add_url_rule("/preferences", methods=['GET'], view_func=ShortConnHandler.as_view("download_preferences"))
        app.add_url_rule("/preferences", methods=['PATCH'], view_func=ShortConnHandler.as_view("edit_preferences"))
        app.add_url_rule("/preferences", methods=['DELETE'], view_func=ShortConnHandler.as_view("delete_preferences"))
        app.add_url_rule("/preferences", methods=['POST'], view_func=ShortConnHandler.as_view("reset_preferences"))
        app.add_url_rule("/register", methods=['GET'], view_func=ShortConnHandler.as_view("download_token", val=False))
        app.add_url_rule("/legality", methods=['GET'], view_func=ShortConnHandler.as_view("check_legality"))
        app.add_url_rule("/emotion", methods=['GET'], view_func=ShortConnHandler.as_view("normalize_emo"))
        app.add_url_rule("/vista", methods=['POST'], view_func=ShortConnHandler.as_view("upload_vista"))
        app.add_url_rule("/vista", methods=['DELETE'], view_func=ShortConnHandler.as_view("delete_vista"))
        app.add_url_rule("/vista", methods=['GET'], view_func=ShortConnHandler.as_view("download_vista"))
        app.add_url_rule("/servers", methods=['GET'], view_func=ShortConnHandler.as_view("get_servers", val=False))
        app.add_url_rule("/accessibility", methods=['GET'], view_func=ShortConnHandler.as_view("get_accessibility", val=False))
        app.add_url_rule("/version", methods=['GET'], view_func=ShortConnHandler.as_view("get_version", val=False))
        app.add_url_rule("/workload", methods=['GET'], view_func=ShortConnHandler.as_view("get_workload", val=False))
        app.add_url_rule("/defaults", methods=['GET'], view_func=ShortConnHandler.as_view("get_defaults", val=False))
    else:
        app.add_url_rule("/savefile", methods=['POST'], view_func=ShortConnHandler.as_view("upload_savefile"))
        app.add_url_rule("/savefile/delete", methods=['POST'], view_func=ShortConnHandler.as_view("delete_savefile"))
        app.add_url_rule("/trigger", methods=['POST'], view_func=ShortConnHandler.as_view("upload_trigger"))
        app.add_url_rule("/trigger/delete", methods=['POST'], view_func=ShortConnHandler.as_view("delete_trigger"))
        app.add_url_rule("/history", methods=['GET'], view_func=ShortConnHandler.as_view("download_history"))
        app.add_url_rule("/history", methods=['POST'], view_func=ShortConnHandler.as_view("restore_history"))
        app.add_url_rule("/preferences", methods=['GET'], view_func=ShortConnHandler.as_view("download_preferences"))
        app.add_url_rule("/preferences/edit", methods=['POST'], view_func=ShortConnHandler.as_view("edit_preferences"))
        app.add_url_rule("/preferences/delete", methods=['POST'], view_func=ShortConnHandler.as_view("delete_preferences"))
        app.add_url_rule("/preferences/reset", methods=['POST'], view_func=ShortConnHandler.as_view("reset_preferences"))
        app.add_url_rule("/register", methods=['GET'], view_func=ShortConnHandler.as_view("download_token", val=False))
        app.add_url_rule("/legality", methods=['GET'], view_func=ShortConnHandler.as_view("check_legality"))
        app.add_url_rule("/emotion", methods=['GET'], view_func=ShortConnHandler.as_view("normalize_emo"))
        app.add_url_rule("/vista", methods=['POST'], view_func=ShortConnHandler.as_view("upload_vista"))
        app.add_url_rule("/vista/delete", methods=['POST'], view_func=ShortConnHandler.as_view("delete_vista"))
        app.add_url_rule("/vista", methods=['GET'], view_func=ShortConnHandler.as_view("download_vista"))
        app.add_url_rule("/servers", methods=['GET'], view_func=ShortConnHandler.as_view("get_servers", val=False))
        app.add_url_rule("/accessibility", methods=['GET'], view_func=ShortConnHandler.as_view("get_accessibility", val=False))
        app.add_url_rule("/version", methods=['GET'], view_func=ShortConnHandler.as_view("get_version", val=False))
        app.add_url_rule("/workload", methods=['GET'], view_func=ShortConnHandler.as_view("get_workload", val=False))
        app.add_url_rule("/defaults", methods=['GET'], view_func=ShortConnHandler.as_view("get_defaults", val=False))
    app.add_url_rule("/<path>", methods=['GET', 'POST', 'PUT', 'DELETE', 'PATCH'], view_func=ShortConnHandler.as_view("any_unknown", val=False))

    try:
        KNOWN_SERVERS = json.loads(G.A.SERVERS_LIST)
    except Exception as e:
        sync_messenger(info=f'Loading servers list failed: {str(e)}', type=MsgType.ERROR)
        KNOWN_SERVERS = False

app = Quart(import_name=__name__)
app.config['JSON_AS_ASCII'] = False
app.config['MAX_CONTENT_LENGTH'] = 32 * 1024 * 1024

quart_logger = logging.getLogger('hypercorn.error')
quart_logger.disabled = True

class ShortConnHandler(View):
    """Flask initiates it on every request."""
    init_every_request = True
    stem_inst: Optional[NoWsCoroutine]

    root_csc: ConnSocketsContainer = None
    """Don't forget to implement at first!"""

    nvwatchers: list[NvWatcher] = []

    def __init__(self, val=True):
        self.val = val
        if val:
            rsc = RealtimeSocketsContainer()
            csc = self.__class__.root_csc.spawn_sub(rsc) if self.__class__.root_csc else ConnSocketsContainer()
            self.fsc = FullSocketsContainer(rsc, csc)
            
            self.maica_pool = self.fsc.maica_pool
            self.remote_addr = None

    def msg_http(self, *args, **kwargs):
        if self.val:
            sync_messenger(*args, **kwargs)

    @staticmethod
    def jfy_res(content=None):
        """I mean jsonify result."""
        d = {"success": True, "exception": None}
        if content:
            d['content'] = content
        return jsonify(d)

    async def dispatch_request(self, **kwargs):
        try:
            if self.val:
                self.stem_inst = await NoWsCoroutine.async_create(self.fsc)
                self.settings = self.fsc.maica_settings
            else:
                self.stem_inst = None
                self.settings = None
            endpoint = request.endpoint
            function_routed = getattr(self, endpoint)

            xff = request.headers.get('X-Forwarded-For')
            if xff:
                self.remote_addr = xff.split(',')[0].strip()
                if self.stem_inst:
                    self.stem_inst.remote_addr = self.remote_addr

            self.msg_http(info=f'Recieved request on API endpoint {endpoint}', type=MsgType.RECV)
            self.msg_http(info=f'From IP {self.remote_addr}', type=MsgType.DEBUG)
            result = await function_routed()

            if isinstance(result, Response):
                result_json = await result.get_json()
                if result_json:
                    d = {"success": result_json.get('success'), "exception": result_json.get('exception')}
                    if "content" in result_json:
                        d["content"] = ellipsis_str(result_json.get('content'), limit=65)
                    self.msg_http(info=f'Return value: {str(d)}', type=MsgType.SYS)
                else:
                    self.msg_http(info='A non-json response has been made', type=MsgType.SYS)

            return result

        except CommonMaicaException as ce:
            if ce.is_critical:
                traceback.print_exc()
            await messenger(error=ce, no_raise=True)
            return jsonify({"success": False, "exception": str(ce)}), int(ce.error_code) or 400

        except Exception as e:
            await messenger(info=f'Handler hit an exception: {str(e)}', type=MsgType.WARN)
            return jsonify({"success": False, "exception": str(e)}), 400

    async def validate_http(self, raw_data: Union[str, dict], must: Optional[list]=None) -> dict:
        must = must or []
        data_json = await validate_input(raw_data, 100000, None, must=must)
        if self.val and 'access_token' in must:
            access_token = data_json.get('access_token')
            assert access_token, "access_token not provided"
            login_result = await self.stem_inst.hash_and_login(access_token)
            assert login_result, "Login failed somehow"

        if 'chat_session' in must:
            data_json['chat_session'] = int(data_json['chat_session'])
            assert 0 <= data_json.get('chat_session') < 10, "chat_session out of bound"

        return data_json

    async def upload_savefile(self):
        """POST"""
        json_data = await request.get_json()
        valid_data = await self.validate_http(json_data, must=['access_token', 'chat_session', 'content'])

        chat_session = valid_data.get('chat_session')
        content = valid_data.get('content')
        content_str = json.dumps(valid_data.get('content'), ensure_ascii=False) if content else None

        sql_expression_1 = "SELECT persistent_id FROM persistents WHERE user_id = %s AND chat_session_num = %s"
        result = await self.maica_pool.query_get(sql_expression_1, (self.settings.verification.user_id, chat_session))
        if result:
            persistent_id = result[0]
            sql_expression_2 = "UPDATE persistents SET content = %s WHERE persistent_id = %s"
            await self.maica_pool.query_modify(sql_expression_2, (content_str, persistent_id))
        else:
            sql_expression_2 = "INSERT INTO persistents (user_id, chat_session_num, content) VALUES (%s, %s, %s)"
            await self.maica_pool.query_modify(sql_expression_2, (self.settings.verification.user_id, chat_session, content_str))

        return self.jfy_res()
    
    async def delete_savefile(self):
        """DELETE"""
        json_data = await request.get_json()
        valid_data = await self.validate_http(json_data, must=['access_token', 'chat_session'])

        chat_session = valid_data.get('chat_session')

        sql_expression_1 = "DELETE FROM persistents WHERE user_id = %s AND chat_session_num = %s"
        await self.maica_pool.query_modify(sql_expression_1, (self.settings.verification.user_id, chat_session))

        return self.jfy_res()
        
    async def upload_trigger(self):
        """POST"""
        json_data = await request.get_json()
        valid_data = await self.validate_http(json_data, must=['access_token', 'chat_session', 'content'])

        chat_session = valid_data.get('chat_session')
        content = valid_data.get('content')
        content_str = json.dumps(valid_data.get('content'), ensure_ascii=False) if content else None

        sql_expression_1 = "SELECT trigger_id FROM triggers WHERE user_id = %s AND chat_session_num = %s"
        result = await self.maica_pool.query_get(sql_expression_1, (self.settings.verification.user_id, chat_session))
        if result:
            trigger_id = result[0]
            sql_expression_2 = "UPDATE triggers SET content = %s WHERE trigger_id = %s"
            await self.maica_pool.query_modify(sql_expression_2, (content_str, trigger_id))
        else:
            sql_expression_2 = "INSERT INTO triggers (user_id, chat_session_num, content) VALUES (%s, %s, %s)"
            await self.maica_pool.query_modify(sql_expression_2, (self.settings.verification.user_id, chat_session, content_str))

        return self.jfy_res()
    
    async def delete_trigger(self):
        """DELETE"""
        json_data = await request.get_json()
        valid_data = await self.validate_http(json_data, must=['access_token', 'chat_session'])

        chat_session = valid_data.get('chat_session')

        sql_expression_1 = "DELETE FROM triggers WHERE user_id = %s AND chat_session_num = %s"
        await self.maica_pool.query_modify(sql_expression_1, (self.settings.verification.user_id, chat_session))

        return self.jfy_res()
   
    async def download_history(self):
        """GET"""
        json_data = request.args.to_dict(flat=True)
        valid_data = await self.validate_http(json_data, must=['access_token', 'chat_session'])

        chat_session = valid_data.get('chat_session')
        assert 1 <= chat_session < 10, "chat_session out of bound"
        rounds = int(valid_data.get('content', 0))

        history_json = (await self.stem_inst.rw_chat_session('r', chat_session_num=chat_session))[1]

        if history_json:
            history_len = len(history_json)
            # If required length larger than overall length
            if abs(2 * int(rounds)) + 1 >= history_len:
                history_final_json = history_json
            else:
                match int(rounds):
                    case i if i > 0:
                        history_final_json = history_json[:(2 * i + 1)]
                    case i if i < 0:
                        history_final_json = [history_json[0]] + history_json[(2 * i):]
                    case _:
                        history_final_json = history_json
            history_final_str = json.dumps(history_final_json, ensure_ascii=False, sort_keys=True)
            sigb64 = await wrap_run_in_exc(None, sign_message, history_final_str)
            return self.jfy_res([sigb64, history_final_json])
        else:
            return self.jfy_res([])

    async def restore_history(self):
        """PUT"""
        json_data = await request.get_json()
        valid_data = await self.validate_http(json_data, must=['access_token', 'chat_session', 'content'])

        chat_session = valid_data.get('chat_session')
        assert 1 <= chat_session < 10, "chat_session out of bound"
        content = valid_data.get('content')

        sigb64, history_json = content
        assert (await wrap_run_in_exc(None, verify_message, json.dumps(history_json, ensure_ascii=False, sort_keys=True), sigb64)), "Signature mismatch"
        await self.stem_inst.restore_chat_session(history_json, chat_session)

        return self.jfy_res()

    async def download_preferences(self):
        """GET"""
        json_data = request.args.to_dict(flat=True)
        valid_data = await self.validate_http(json_data, must=['access_token'])

        preferences_json = await self.stem_inst.hasher.check_user_status(pref=True)

        return self.jfy_res(preferences_json)

    async def edit_preferences(self):
        """PATCH"""
        json_data = await request.get_json()
        valid_data = await self.validate_http(json_data, must=['access_token', 'content'])

        preferences_json = await self.stem_inst.hasher.check_user_status(pref=True)

        content = valid_data.get('content')
        assert isinstance(content, dict), "Request content invalid" 
        preferences_json.update(content)
        await self.validate_http(preferences_json)

        await self.stem_inst.hasher.write_user_status(enforce=True, pref=True, **preferences_json)
        return self.jfy_res()

    async def delete_preferences(self):
        """DELETE"""
        json_data = await request.get_json()
        valid_data = await self.validate_http(json_data, must=['access_token', 'content'])

        preferences_json = await self.stem_inst.hasher.check_user_status(pref=True)

        content = valid_data.get('content')
        assert isinstance(content, list), "Request content invalid" 
        for key in content:
            preferences_json.pop(key)

        await self.stem_inst.hasher.write_user_status(enforce=True, pref=True, **preferences_json)
        return self.jfy_res()

    async def reset_preferences(self):
        """POST"""
        json_data = await request.get_json()
        valid_data = await self.validate_http(json_data, must=['access_token'])

        await self.stem_inst.hasher.write_user_status(enforce=True, pref=True)
        return self.jfy_res()
        
    # async def control_preferences(self):
    #     """read->GET, write->PATCH, delete->DELETE, reset->POST"""
    #     try:
    #         json_data = request.args.to_dict(flat=True) if request.method == 'GET' else await request.get_json()
    #         must = ['access_token'] if request.method in ['GET', 'POST'] else ['access_token', 'content']
    #         valid_data = await self.validate_http(json_data, must=must)

    #         if request.method == 'POST':
    #             await self.stem_inst.hasher.write_user_status(True, True)
    #             return jsonify({"success": True, "exception": None})
    #         else:
    #             preferences_json = await self.stem_inst.hasher.check_user_status(True)

    #             if request.method == 'GET':
    #                 return jsonify({"success": True, "exception": None, "content": preferences_json})
    #             else:
    #                 content = valid_data.get('content')
    #                 if request.method == 'PATCH':
    #                     assert isinstance(content, dict), "Request content invalid"
    #                     preferences_json.update(content)
    #                 elif request.method == 'DELETE':
    #                     assert isinstance(content, list), "Request content invalid"
    #                     for key in content:
    #                         preferences_json.pop(key)
    #                 await self.stem_inst.hasher.write_user_status(True, True, **preferences_json)
    #                 return jsonify({"success": True, "exception": None})
                
    #     except Exception as e:
    #         return jsonify({"success": False, "exception": str(e)})

    async def download_token(self):
        """GET, val=False"""
        json_data = request.args.to_dict(flat=True)
        valid_data = await self.validate_http(json_data, must=['content'])

        content = json.loads(valid_data.get('content'))

        cridential_type = 'username' if content.get('username') else 'email'
        cridential = content.get(cridential_type)
        password = content.get('password')
        assert isinstance(cridential, str) and isinstance(password, str), "Request content invalid"

        unencrypted_token = json.dumps({cridential_type: cridential, "password": password}, ensure_ascii=False)
        encrypted_token = await wrap_run_in_exc(None, encrypt_token, unencrypted_token)
        return self.jfy_res(encrypted_token)

    async def check_legality(self):
        """GET"""
        json_data = request.args.to_dict(flat=True)
        valid_data = await self.validate_http(json_data, must=['access_token'])

        try:
            content = json.loads(valid_data.get('content'))
        except:
            content = None

        if not content:
            result = self.settings.verification.username
        else:
            object = content.get('object')
            value = content.get('value')
            assert object and value, 'Empty in input'
            match object:
                case 'geolocation':
                    res = await weather_api_get(value)
                    result = res

                case _:
                    raise MaicaInputWarning(f"'{object}' is not valid object")

        return self.jfy_res(result)

    async def normalize_emo(self):
        """GET"""
        json_data = request.args.to_dict(flat=True)
        valid_data = await self.validate_http(json_data, must=['access_token', 'content'])

        content = json.loads(valid_data.get('content'))

        proc_type: Literal['norm', 'add'] = content.get('type') or 'norm'
        proc_lang: Literal['zh', 'en'] = content.get('target_lang') or 'zh'
        emo = content.get('text')
        
        if proc_type == 'norm':
            result = await emo_proc_auto(emo, proc_lang, self.fsc.mnerve_conn)
        else:
            result = await emo_proc_llm(emo, proc_lang, self.fsc.mnerve_conn)

        return self.jfy_res(result)

    async def upload_vista(self):
        """POST, multipart/form-data"""
        json_data = (await request.form).to_dict()
        valid_data = await self.validate_http(json_data, must=['access_token'])

        file: FileStorage = (await request.files).get('content')

        img_uuid = await self.stem_inst.store_mv(file.stream.read())

        return self.jfy_res(img_uuid)
    
    async def delete_vista(self):
        """DELETE"""
        json_data = await request.get_json()
        valid_data = await self.validate_http(json_data, must=['access_token'])

        content = valid_data.get('content')

        await self.stem_inst.delete_mv(content)
        return self.jfy_res()
    
    async def download_vista(self):
        """GET"""
        json_data = request.args.to_dict(flat=True)
        try:
            # Predicted trying to download img
            valid_data = await self.validate_http(json_data, must=['content'])
            content = valid_data.get('content')

            uuid: str = content
            processing_img = ProcessingImg(uuid)

            return_bio = processing_img.to_bio()
            file_name = processing_img.file_name

            return await send_file(
                return_bio,
                as_attachment=True,
                attachment_filename=file_name
            )
        except Exception as e:
            # Trying to get img list
            try:
                valid_data = await self.validate_http(json_data, must=['access_token'])
                result = await self.stem_inst.list_user_mv()
                return self.jfy_res(result)
            except Exception as e2:
                if isinstance(valid_data, dict) and valid_data.get('content'):
                    raise MaicaInputWarning(f"File {valid_data.get('content')}.jpg not exist", '404') from e
                else:
                    raise e2

    async def get_servers(self):
        """GET, val=False"""
        return self.jfy_res(KNOWN_SERVERS)
    
    async def get_accessibility(self):
        """GET, val=False"""
        accessibility = G.A.DEV_STATUS
        return self.jfy_res(accessibility)
    
    async def get_version(self):
        """GET, val=False"""
        curr_version, legc_version = G.A.CURR_VERSION, G.A.LEGC_VERSION
        blessland_capv = G.A.BLESSLAND_CAPV
        return self.jfy_res({"curr_version": curr_version, "legc_version": legc_version, "fe_blessland_version": blessland_capv})

    async def get_workload(self):
        """GET, val=False"""
        content = {}
        for watcher in self.__class__.nvwatchers:
            content |= watcher.get_statics_inside()

        content.update({"onliners": len(online_dict)})
        return self.jfy_res(content)
    
    async def get_defaults(self):
        """GET, val=False"""
        settings = MaicaSettings()
        content = {}
        content.update(settings.basic.default())
        content.update(settings.extra.default())
        content.update(settings.super.default())
        return self.jfy_res(content)

    async def any_unknown(self):
        """Handles any unknown endpoint"""
        await messenger(info=f"An unknown access to {request.full_path} handled", type=MsgType.WARN)
        return jsonify({"success": False, "exception": 'Unknown request endpoint or method'}), 404

async def prepare_thread(**kwargs):

    # Construct csc first
    root_csc_kwargs = {k: kwargs.get(k) for k in _CONNS_LIST}
    root_csc = ConnSocketsContainer(**root_csc_kwargs)
    ShortConnHandler.root_csc = root_csc

    await messenger(info='MAICA HTTP server started!', type=MsgType.PRIM_SYS)

    # Construct watchers
    watch_addrs = {}
    for k, v in _WATCHES_DICT.items():
        host = get_host(getattr(G.A, v))
        if host and not k in watch_addrs:
            watch_addrs[k] = host
            
    _watch_start_list = []
    for k, _ in watch_addrs.items():
        watcher = await NvWatcher.async_create(k, 'maica')
        ShortConnHandler.nvwatchers.append(watcher)
        _watch_start_list.append(asyncio.create_task(watcher.wrapped_main_watcher()))

    try:
        config = Config()
        config.bind = ['0.0.0.0:6000']
        task = asyncio.create_task(serve(app, config))
        task_list = [task] + _watch_start_list
        await asyncio.wait(task_list, return_when=asyncio.FIRST_COMPLETED)
    except BaseException as be:
        if isinstance(be, Exception):
            error = CommonMaicaError(str(be), '504')
            await messenger(error=error, no_raise=True)
    finally:

        # Normally maica_http should be the first one (possibly only one) to
        # respond to the original SIGINT.

        # So its stop msg will be print first, adding \n after ^C to look prettier.

        await messenger(info='\n', type=MsgType.PLAIN)
        await messenger(info='MAICA HTTP server stopped!', type=MsgType.PRIM_SYS)

async def _run_http():
    from maica import init
    init()
    pkg_init_maica_http()
    _root_csc_items = [getattr(ConnUtils, k)() for k in _CONNS_LIST]
    root_csc_items = await asyncio.gather(*_root_csc_items)
    root_csc_kwargs = dict(zip(_CONNS_LIST, root_csc_items))

    task = asyncio.create_task(prepare_thread(**root_csc_kwargs))
    await task

    close_list = []
    for conn in root_csc_items:
        close_list.append(conn.close())
    await asyncio.gather(*close_list)
    await messenger(info='Individual MAICA HTTP server cleaning done', type=MsgType.DEBUG)

def run_http(**kwargs):
    asyncio.run(_run_http())

if __name__ == '__main__':
    run_http()


import os
import re
from maica.maica_utils import load_env, get_inner_path, sync_messenger, MsgType

mark_path = get_inner_path('.initialized')

def check_marking():
    init_version_line = re.compile(r'Last migrated version:\s*(.*?)\s*$')
    try:
        with open(mark_path, 'r') as mark_file:
            lines = mark_file.readlines()
            last_version = '1.0.0000'
        for line in lines:
            res = init_version_line.match(line)
            if res:
                last_version = res[1]
                break
        else:
            sync_messenger(info='Failed to read last migrated version from .initialized, will assume <=1.1.007.post2', type=MsgType.WARN)
        return last_version

    except:
        return False
    
def create_marking():
    curr_version, legc_version = load_env('MAICA_CURR_VERSION'), load_env('MAICA_VERSION_CONTROL')
    with open(mark_path, 'w') as mark:
        mark.write(f"This file's existence indicates that the program has been initialized once.\nTo try initializing it again, delete this file.\n\nWarning: Deleting this file will NOT make the program run any cleanups.\n\nLast migrated version: {curr_version}")
    return

if __name__ == "__main__":
    print(check_marking())
    create_marking()
    print(check_marking())


import asyncio
import os
from typing import *
from maica.maica_utils import *

upper_version = "1.1.007.post3"

async def migrate():

    maica_pool = await ConnUtils.maica_pool()
    try:
        await maica_pool.query_modify('RENAME TABLE cchop_archived TO crop_archived')
    except Exception as e:
        raise MaicaDbWarning(f'Couldn\'t rename table cchop_archived to crop_archived: {str(e)}, maybe manually done already?') from e
    finally:
        await maica_pool.close()



import asyncio
import os
from typing import *
from maica.maica_utils import *

upper_version = "1.2.003"

async def migrate():

    auth_pool = await ConnUtils.auth_pool(ro=False)
    try:
        if auth_pool.db_type == 'mysql':
            result = await auth_pool.query_get("select count(*) from information_schema.columns where table_name = 'users' and column_name = 'suspended_until'")
            exist = result[0]
            if not exist:
                sync_messenger(info='Column does not exist, creating...', type=MsgType.DEBUG)
                await auth_pool.query_modify('alter table `users` add column `suspended_until` datetime DEFAULT NULL')
            else:
                sync_messenger(info='Column already exists', type=MsgType.DEBUG)
        else:
            result = await auth_pool.query_get(
                "SELECT name FROM pragma_table_info('users') WHERE name='suspended_until'"
            )
            if not result:
                sync_messenger(info='Column does not exist, creating...', type=MsgType.DEBUG)
                await auth_pool.query_modify(
                    'ALTER TABLE users ADD COLUMN suspended_until TIMESTAMP'
                )
            else:
                sync_messenger(info='Column already exists', type=MsgType.DEBUG)
    except Exception as e:
        raise MaicaDbWarning(f'Couldn\'t add column \'suspended_until\' automatically: {str(e)}, consider doing a manual double-check') from e
    finally:
        await auth_pool.close()


import json
import datetime
import asyncio
from maica.maica_utils import *
from .api_keys import TpAPIKeys

async def weather_api_get(location):
    try:
        wkey = TpAPIKeys.GAODE_WEATHER
        d = {
            "geocode": None,
            "temperature": None,
            "weather": None,
            "humidity": None,
        }
        addr_code = (await dld_json(f"https://restapi.amap.com/v3/geocode/geo?key={wkey}&address={location}"))['geocodes'][0]['adcode']
        result = (await dld_json(f"https://restapi.amap.com/v3/weather/weatherInfo?key={wkey}&city={addr_code}&extensions=base"))['lives'][0]
        d['geocode'], d['temperature'], d['weather'], d['humidity'] = addr_code, result['temperature'], result['weather'], result['humidity']
        return d
    
    except CommonMaicaException as ce:
        raise ce
    
    except Exception as e:
        raise MaicaInternetWarning(f'Weather API failed: {str(e)}', '406') from e

if __name__ == '__main__':
    from maica import init
    init()
    print(asyncio.run(weather_api_get('los angeles')))



"""
This is third party (https://app.scrapeless.com/) implementation.
Works okay, decent price, I know nothing else.
"""

import asyncio
import httpx
import json

from typing import *
from maica.maica_utils import *
from maica.mtools.api_keys import TpAPIKeys

requires = ['SCRAPELESS_SERP']
async def asearch(query, target_lang: Literal['zh', 'en']='zh'):
    host = "api.scrapeless.com"
    url = f"https://{host}/api/v1/scraper/request"
    token = TpAPIKeys.SCRAPELESS_SERP

    headers = {
        "x-api-token": token,
        "Content-Type": "application/json"
    }
    json_payload = json.dumps({
        "actor": "scraper.google.search",
        "input": {
            "q": query,
            "gl": "cn" if target_lang == 'zh' else 'us',
            "hl": "zh" if target_lang == 'zh' else 'en',
            "google_domain": "google.com",
            "num": 20,
        },
        "waiting": True
    })

    async with httpx.AsyncClient(proxy=G.A.PROXY_ADDR) as client:
        response = await client.post(url, headers=headers, data=json_payload, timeout=30)
        response_json = response.json()

    results_formatted = [{"title": it['title'], "text": it['snippet']} for it in response_json['organic_results'] if 'snippet' in it]
    return results_formatted

if __name__ == "__main__":
    async def main():
        from maica import init
        init()
        await asearch("今天的新闻")

    asyncio.run(main())
    


import asyncio
import json

from mcp.client.stdio import stdio_client
from mcp import ClientSession, StdioServerParameters
from .mcp_middleware import prep_bin, _serp_bin
from maica.maica_utils import *

def pkg_init_mcp_serp():
    global SERP_BIN
    SERP_BIN = _serp_bin()
    prep_bin(SERP_BIN)

async def asearch(query, target_lang='zh'):
    """
    Async google searcher with (somewhat) MCP protocol.
    Uses an external precompiled NodeJS project to scrape, refer to https://github.com/edgeinfinity1/mi-search-mcp
    """
    locale = 'zh-CN' if target_lang == 'zh' else 'en-US'

    serp_initiation_args = {
        "command": SERP_BIN,
        "args": [],
        "env": {k: G.A.PROXY_ADDR for k in ['http_proxy', 'HTTP_PROXY', 'https_proxy', 'HTTPS_PROXY']},
    }
    serp_initiation_args["env"].update({
        "DISPLAY": (load_env("DISPLAY") or '') if G.A.ENABLE_X11 == '1' else ''
    })

    serp_initiation = StdioServerParameters(**serp_initiation_args)

    async with stdio_client(serp_initiation) as (stdio, write):
        async with ClientSession(stdio, write) as session:
            await session.initialize()

            # tools = await session.list_tools():
            response = await session.call_tool('search', {'queries': [query], "limit": 10, "locale": locale, "debug": False})
            response_json = json.loads(response.content[0].text)
            return [{"title": it['title'], "text": it['snippet']} for it in response_json['searches'][0]['results']]

if __name__ == '__main__':
    from maica import init
    init()
    pkg_init_mcp_serp()
    print(asyncio.run(asearch("测试")))


import asyncio
import os
from typing import *
from maica.maica_utils import *

upper_version = "1.2.000.rc2"

async def migrate():

    maica_pool = await ConnUtils.maica_pool()
    try:
        if maica_pool.db_type == 'mysql':
            await maica_pool.query_modify('ALTER TABLE `account_status` CHANGE `status` `status` TEXT CHARACTER SET utf8mb4 COLLATE utf8mb4_general_ci NULL DEFAULT NULL; ')
            await maica_pool.query_modify('ALTER TABLE `account_status` CHANGE `preferences` `preferences` TEXT CHARACTER SET utf8mb4 COLLATE utf8mb4_general_ci NULL DEFAULT NULL; ')
            await maica_pool.query_modify('ALTER TABLE `ms_cache` CHANGE `hash` `hash` TEXT CHARACTER SET utf8mb4 COLLATE utf8mb4_general_ci NOT NULL; ')
            await maica_pool.query_modify('ALTER TABLE `ms_cache` CHANGE `content` `content` TEXT CHARACTER SET utf8mb4 COLLATE utf8mb4_general_ci NULL DEFAULT NULL; ')
            await maica_pool.query_modify("""
CREATE TABLE IF NOT EXISTS `mv_meta` (
  `vista_id` int(11) NOT NULL AUTO_INCREMENT,
  `user_id` int(11) NOT NULL,
  `uuid` text NOT NULL,
  `timestamp` datetime NOT NULL DEFAULT CURRENT_TIMESTAMP ON UPDATE CURRENT_TIMESTAMP,
  PRIMARY KEY (`vista_id`)
)
                                        """)
        else:
            await maica_pool.query_modify("""
CREATE TABLE IF NOT EXISTS mv_meta (
  vista_id INTEGER PRIMARY KEY AUTOINCREMENT,
  user_id INTEGER NOT NULL,
  uuid TEXT NOT NULL,
  timestamp DATETIME DEFAULT CURRENT_TIMESTAMP
)
                                        """)
            await maica_pool.query_modify("""
CREATE TRIGGER IF NOT EXISTS update_timestamp_mv_meta
AFTER UPDATE ON mv_meta
FOR EACH ROW
BEGIN
    UPDATE mv_meta
    SET timestamp = CURRENT_TIMESTAMP
    WHERE vista_id = OLD.vista_id;
END;
                                        """)

    except Exception as e:
        raise MaicaDbWarning(f'Couldn\'t alter lines LONGTEXT to TEXT and add table mv_meta: {str(e)}, maybe manually done already?') from e
    finally:
        await maica_pool.close()



import os
import sys
import traceback
from typing import *
from pathlib import Path
from dotenv import load_dotenv
try:
    from maica import maica_ws
except Exception:
    current_dir = Path(__file__).parent
    parent_dir = current_dir.parent
    if str(parent_dir) not in sys.path:
        sys.path.insert(0, str(parent_dir))

try:
    import mtts
    from mtts.mtts_utils import locater as mtts_locater
    mtts_installed = True
except Exception:
    mtts_installed = False

import asyncio
import argparse
import colorama

from maica import maica_ws, maica_http, common_schedule, silent as _silent
from maica.maica_utils import *
from maica.initializer import *

_CHAT_CONNS_LIST = ['auth_pool', 'maica_pool', 'mcore_conn', 'mfocus_conn', 'mvista_conn', 'mnerve_conn']
_TTS_CONNS_LIST = ['auth_pool', 'maica_pool']

from maica.initializer import pkg_init_initializer
from maica.maica_http import pkg_init_maica_http
from maica.maica_utils import pkg_init_maica_utils
from maica.mtools import pkg_init_mtools
def pkg_init_maica():
    pkg_init_initializer()
    """Prio 0"""
    pkg_init_maica_utils()
    """Prio 1"""
    pkg_init_maica_http()
    pkg_init_mtools()
    if mtts_installed:
        mtts.mtts_http.pkg_init_mtts_http()

colorama.init(autoreset=True)
initialized = False
start_target: Literal['chat', 'tts', 'all'] = 'chat'

def check_params(envdir: str=None, extra_envdir: list=None, silent=False, **kwargs):
    """This will only run once. Recalling will not take effect, except passing in extra kwargs."""
    global initialized, start_target

    def init_parser():
        parser = argparse.ArgumentParser(description="Start MAICA Illuminator deployment or run self-maintenance functions")
        parser.add_argument('target', choices=['chat', 'tts', 'all'], nargs='?', default='chat', help='Start maica chat server or mtts server, default chat')
        parser.add_argument('-e', '--envdir', help='Include external env file for running deployment, specify every time')
        parser.add_argument('-s', '--silent', action="store_true", help='Run without logging (unrecommended for deployment)')
        excluse_1 = parser.add_mutually_exclusive_group()
        excluse_1.add_argument('-k', '--keys', choices=['path', 'export', 'import'], nargs='?', const='path', help='Get path of RSA keys, or export/import to/from current directory')
        excluse_1.add_argument('-d', '--databases', choices=['path', 'export', 'import'], nargs='?', const='path', help='Get path of databases, or export/import to/from current directory')
        excluse_1.add_argument('-t', '--templates', choices=['print', 'create'], nargs='?', const='print', help='Print config templates or create them in current directory')
        return parser
    
    parser = init_parser()
    args = parser.parse_args()
    start_target = args.target
    envdir = envdir or args.envdir
    silent = silent or args.silent
    operate_keys = args.keys
    operate_databases = args.databases
    operate_templates = args.templates

    _silent(silent)

    def dest_env(envdir: str=None, extra_envdir: list=None):
        """This part we load all static env vars."""
        if not envdir:
            realpath = get_inner_path('.env')
            sync_messenger(info=f'[maica-env] No env file designated, defaulting to {realpath}...', type=MsgType.DEBUG)
            envdir = realpath
            if not os.path.isfile(envdir):
                realpath = os.path.abspath('.env')
                sync_messenger(info=f'[maica-env] {envdir} is not a file, trying {realpath}...', type=MsgType.DEBUG)
        else:
            realpath = os.path.abspath(envdir)

        if os.path.isfile(realpath):
            sync_messenger(info=f'[maica-env] Loading env file {realpath}...', type=MsgType.DEBUG)
            load_dotenv(dotenv_path=realpath)
        else:
            sync_messenger(info=f'[maica-env] {realpath} is not a file, skipping real env file...', type=MsgType.WARN)

        if extra_envdir:
            for edir in extra_envdir:
                realpath = os.path.abspath(edir)
                if os.path.isfile(realpath):
                    sync_messenger(info=f'[maica-env] Loading extra env file {realpath}...', type=MsgType.DEBUG)
                    load_dotenv(dotenv_path=realpath)

        if mtts_installed:
            realpath = mtts_locater.get_inner_path('mtts_env_basis')
            sync_messenger(info=f'[maica-env] Loading mtts basis {realpath} to guarantee basic functions...', type=MsgType.DEBUG)
            if os.path.isfile(realpath):
                load_dotenv(dotenv_path=realpath)
            else:
                raise Exception('env basis lost')

        realpath = get_inner_path('env_basis')
        sync_messenger(info=f'[maica-env] Loading maica basis {realpath} to guarantee basic functions...', type=MsgType.DEBUG)
        if os.path.isfile(realpath):
            load_dotenv(dotenv_path=realpath)
        else:
            raise Exception('env basis lost')

    def get_templates():
        with open(get_inner_path('env_basis'), 'r', encoding='utf-8') as env_e:
            env_c = env_e.readlines()
        i = 0; j = 0
        for l in env_c:
            i += 1
            if len(l) <= 5:
                j = 1
            if j and len(l) > 5:
                break
        env_c = ''.join(env_c[i - 1:])
        return env_c
    
    def separate_line(title: str):
        try:
            terminal_width = os.get_terminal_size().columns
        except:
            terminal_width = 40
        line = title.center(terminal_width, '/')
        return line
    
    def print_templates():
        env_c = get_templates()
        print(colorama.Fore.BLUE + separate_line('Begin .env template'))
        print(env_c)
        print(colorama.Fore.BLUE + separate_line('End .env template'))

    def print_path(*files: list, common: str="keys"):
        print(colorama.Fore.BLUE + separate_line(f'Begin {common} paths'))
        for file in files:
            print(get_inner_path(file))
        print(colorama.Fore.BLUE + separate_line(f'End {common} paths'))

    def create_templates():
        env_c = get_templates()
        env_p = os.path.abspath('./.env')
        if os.path.exists(env_p):
            sync_messenger(info=f'[maica-cli] Config {env_p} already exists, skipping creation...', type=MsgType.WARN)
        else:
            with open(env_p, 'w', encoding='utf-8') as env_f:
                sync_messenger(info=f'[maica-cli] Generating {env_p}...', type=MsgType.DEBUG)
                env_f.write(env_c)

        sync_messenger(info='[maica-cli] Creation succeeded, edit them yourself and then start with "maica -c .env"', type=MsgType.LOG)

    if kwargs:
        # Load extra env vars
        for k, v in kwargs.items():
            os.environ[k] = v
        sync_messenger(info=f'[maica-env] Added {len(kwargs)} vars to environ.', type=MsgType.DEBUG)

    if not initialized:
        try:
            if operate_keys:
                match operate_keys:
                    case 'path':
                        print_path('pub.key', 'prv.key')
                    case 'export':
                        export_keys('.')
                    case 'import':
                        import_keys('.')
                exit()

            if operate_templates:
                print_templates() if operate_templates == 'print' else create_templates()
                exit()

            dest_env(envdir, extra_envdir)

            if operate_databases:
                match operate_databases:
                    case 'path':
                        if load_env('MAICA_DB_ADDR') == "sqlite":
                            print_path(load_env('MAICA_AUTH_DB'), load_env('MAICA_DATA_DB'), common='databases')
                        else:
                            sync_messenger(info='[maica-cli] MAICA using served databases!', type=MsgType.WARN)
                    case _:
                        sync_messenger(info='\n[maica-cli] Function yet not supported, do it manually.\nNever ask why we made this.', type=MsgType.LOG)
                exit()

            initialized = True
        except Exception as e:
            sync_messenger(info=f'[maica-init] Error: {str(e)}, quitting...', type=MsgType.ERROR)
            exit(1)

def check_data_init():
    last_version = check_marking()
    if not last_version:
        generate_rsa_keys()
        pkg_init_maica()
        if load_env('MAICA_IS_REAL_ENV') == '1':
            asyncio.run(create_tables())
        else:
            print('''No real env detected, is this workflow?
If it is, at least the imports and grammar are good if you see this.
If not:
    If you're running MAICA for deployment, pass in "--envdir path/to/.env".
    If you're developing with MAICA as dependency, call maica.init(envdir='path/to/.env') after import.
    If you really want to use without manual configuration, call maica.init(ignore_envc=True) after import.
    Or, you can manually set the necessary env vars.
Quitting...'''
                )
            quit(0)
        create_marking()
        sync_messenger(info="[maica-init] MAICA Illuminator initialization finished", type=MsgType.PRIM_SYS)
    else:
        pkg_init_maica()
        sync_messenger(info="[maica-init] Initiated marking detected, checking migrations...", type=MsgType.DEBUG)
    migrated = migrate(last_version)
    if migrated:
        create_marking()

def check_warns():
    if load_env('MAICA_DB_ADDR') == 'sqlite':
        sync_messenger(info='\nMAICA using SQLite database detected\nWhile MAICA Illuminator is fully compatible with SQLite, it can be a significant drawback of performance under heavy workload\nIf this is a public service instance, it\'s strongly recommended to use MySQL/MariaDB instead', type=MsgType.DEBUG)
    if not load_env('MAICA_PROXY_ADDR'):
        sync_messenger(info='\nInternet proxy absence detected\nMAICA Illuminator needs to access Google and Wikipedia for full functionalities, so you\'ll possibly need a proxy for those\nYou can ignore this message safely if you have direct access to those or have a global proxy set already', type=MsgType.DEBUG)
    if load_env('MAICA_DEV_STATUS') != 'serving':
        sync_messenger(info='\nServer status not serving detected\nWith this announcement taking effect, standard clients will stop further communications with MAICA Illuminator respectively\nYou should set DEV_STATUS to \'serving\' whenever the instance is ready to serve', type=MsgType.DEBUG)
    if (load_env('MAICA_MCORE_NODE') and load_env('MAICA_MCORE_USER') and load_env('MAICA_MCORE_PWD')) or (load_env('MAICA_MFOCUS_NODE') and load_env('MAICA_MFOCUS_USER') and load_env('MAICA_MFOCUS_PWD')):
        sync_messenger(info='\nOne or more NVwatch nodes detected\nNVwatch of MAICA Illuminator will try to collect nvidia-smi outputs through SSH, which can fail the process if SSH not avaliable\nIf SSH of nodes are not accessable or not wanted to be used, delete X_NODE, X_USER, X_PWD accordingly', type=MsgType.DEBUG)

async def start_all(start_target: Literal['chat', 'tts', 'all']='chat'):

    _root_csc_items = [getattr(ConnUtils, k)() for k in (_CHAT_CONNS_LIST if start_target != 'tts' else _TTS_CONNS_LIST)]
    root_csc_items = await asyncio.gather(*_root_csc_items)
    root_csc_kwargs = dict(zip(_CHAT_CONNS_LIST, root_csc_items))

    if start_target == 'chat':
        await maica_start_all(**root_csc_kwargs)
    elif start_target == 'tts':
        await mtts_start_all(**root_csc_kwargs)
    else:
        task_chat = asyncio.create_task(maica_start_all(**root_csc_kwargs))
        task_tts = asyncio.create_task(mtts_start_all(**root_csc_kwargs))
        res = await asyncio.wait([
            task_chat,
            task_tts,
            ], return_when=asyncio.FIRST_COMPLETED)
        await messenger(info="First overall quit collected, quitting other tasks...", type=MsgType.DEBUG)
        for pending in res[1]:
            pending.cancel()
            await pending

    await messenger(info="All quits collected, doing final cleanup...", type=MsgType.DEBUG)

    close_list = []
    for conn in root_csc_items:
        if conn:
            close_list.append(conn.close())
    await asyncio.gather(*close_list)

    await messenger(info="Everything done, bye", type=MsgType.DEBUG)
    quit()

async def maica_start_all(**kwargs):

    task_ws = asyncio.create_task(maica_ws.prepare_thread(**kwargs))
    task_http = asyncio.create_task(maica_http.prepare_thread(**kwargs))
    task_schedule = asyncio.create_task(common_schedule.prepare_thread(**kwargs, involve_chat=True, involve_tts=False))

    res = await asyncio.wait([
        task_ws,
        task_http,
        task_schedule,
    ], return_when=asyncio.FIRST_COMPLETED)

    await messenger(info="First chat quit collected, quitting other chat tasks...", type=MsgType.DEBUG)
    for pending in res[1]:
        pending.cancel()
        await pending
    await messenger(info="All chat quits collected", type=MsgType.DEBUG)
    return

async def mtts_start_all(**kwargs):

    if not mtts_installed:
        sync_messenger(info="Install with mi-mtts or .[mtts] to implement", type=MsgType.ERROR)
        return

    task_tts = asyncio.create_task(mtts.prepare_thread(**kwargs))
    task_schedule = asyncio.create_task(common_schedule.prepare_thread(**kwargs, involve_chat=False, involve_tts=True))

    res = await asyncio.wait([
        task_tts,
        task_schedule,
    ], return_when=asyncio.FIRST_COMPLETED)

    await messenger(info="First tts quit collected, quitting other chat tasks...", type=MsgType.DEBUG)
    for pending in res[1]:
        pending.cancel()
        await pending
    await messenger(info="All tts quits collected", type=MsgType.DEBUG)
    return

def full_start():
    check_params()
    check_data_init()
    check_warns()
    asyncio.run(start_all(start_target))

if __name__ == "__main__":
    full_start()